"""Debug and recovery routines that can be used from HardwareManager.

These routines typically only function when you are connected to
and IOTile Device using a special debug adapter, however in that
case they allow memory dumping and forced reprogramming, including
initial code bootstrapping.
"""

import os.path
import os
import subprocess
import tempfile
from typedargs.annotate import context, docannotate
from iotile.core.exceptions import ArgumentError, ExternalError
from iotile.core.utilities.console import ProgressBar
from iotile.core.utilities.intelhex import IntelHex


@context("DebugManager")
class DebugManager(object):
    """Low level debug operations for development and recovery.

    Args:
        stream (CMDStream): A CMDStream subclass that supports
            debug operations on which enable_debug() has already
            been called.
    """

    def __init__(self, stream):
        self._stream = stream

    @docannotate
    def dump_ram(self, out_path, pause=False):
        """Dump all RAM to a binary file.

        Args:
            out_path (path): The output path at which to save
                the binary core dump.  This core dump consists
                of the current contents of RAM for the device.
            pause (bool): Optional parameter to halt chip operation
                while performing the core dump.  Defaults to False,
                which could cause rapidly changing RAM values to
                be in an inconsistent state, but it noninvasive and
                will not interrupt any device activity.
        """

        ram_contents = self._stream.debug_command('dump_ram', {'halt': pause})

        with open(out_path, "wb") as outfile:
            outfile.write(ram_contents)

    @docannotate
    def flash(self, in_path, file_format=None):
        """Flash a new firmware image to the attached chip.

        This flashing takes place over a debug interface and does not require
        a working bootloader.  In particular, this routien is suitable for
        initial board bring-up of a blank MCU.

        If an explicit format is not passed the following rules are used to
        infer the format based on its file extension.

        elf: ELF file
        hex: intel hex file
        bin: raw binary file assumed to start at address 0

        Args:
            in_path (path): The path to the input firmware image that we wish
                to flash.
            file_format (str): Optional explicit format to use to parse the
                input file.  If this is None, the format is automatically
                inferred from the file suffix.  If given explicitly, you
                should pass 'elf', 'hex' or 'bin'.
        """

        format_map = {
            "elf": self._process_elf,
            "hex": self._process_hex,
            "bin": None
        }

        if file_format is None:
            _root, ext = os.path.splitext(in_path)
            if len(ext) > 0:
                file_format = ext[1:]

        format_handler = format_map.get(file_format)
        if format_handler is None:
            raise ArgumentError("Unknown file format or file extension", file_format=file_format, known_formats=[x for x in format_map if format_map[x] is not None])

        base_addresses, section_data = format_handler(in_path)
        for base_address, data in zip(base_addresses, section_data):
            args = {
                'base_address': base_address,
                'data': data
            }

            progress = ProgressBar("Programming Flash")

            def _progress_callback(finished, total):
                progress.count = total
                progress.progress(finished)

            try:
                progress.start()
                self._stream.debug_command('program_flash', args, _progress_callback)
            finally:
                progress.end()

    @classmethod
    def _process_hex(cls, in_path):
        """This function returns a list of base addresses and a list of the binary data 
        for each segment.
        """
        ihex           = IntelHex(in_path)
        segments       = ihex.segments()
        segments_start = [segment[0] for segment in segments]
        segments_data  = [ihex.tobinarray(start=segment[0], end=segment[1]-1) for segment in segments]

        return segments_start, segments_data

    @classmethod
    def _process_elf(cls, in_path):
        tmp = tempfile.NamedTemporaryFile(delete=False)
        tmp.close()

        try:
            err = subprocess.call(['arm-none-eabi-objcopy', '-O', 'ihex', in_path, tmp.name])
            if err != 0:
                raise ExternalError("Cannot convert elf to binary file", error_code=err,
                                    suggestion="Make sure arm-none-eabi-gcc is installed and in your PATH")

            return cls._process_hex(tmp.name)
        finally:
            if os.path.isfile(tmp.name):
                os.remove(tmp.name)



"""All update scripts are composed of a series of records.

All records must inherit from this base class and implement its required methods.
"""

from __future__ import (print_function, absolute_import, unicode_literals)
import struct
from iotile.core.exceptions import ArgumentError, DataError


# pylint: disable=too-few-public-methods;This is an enum class
class MatchQuality(object):
    """Mnemonics for indicating different levels of record matching."""

    NoMatch = 0
    GenericMatch = 50
    PerfectMatch = 100
    DeferMatch = 1000
    PartialMatch = 1100


class DeferMatching(Exception):
    """An exception to indicate that we should accumulate records and match later."""
    def __init__(self, matching_class, partial=None):
        super(DeferMatching, self).__init__()

        self.matching_class = matching_class
        self.partial_match = partial


class UpdateRecord(object):
    """The base class for all update actions inside of an update script."""

    HEADER_LENGTH = 8

    # A map of record types to potentially matching record classes
    KNOWN_CLASSES = {}
    PLUGINS_LOADED = False

    @classmethod
    def MatchType(cls):
        """Return the record type that this record matches.

        All records must match an 8-bit record type field that is used to
        decode a binary script.  Note that multiple records may match the same
        8-bit record type if they have different levels of specificity.

        Returns:
            int: The single record type that this record matches.
        """

        raise NotImplementedError()

    def encode(self):
        """Encode this record into binary, suitable for embedded into an update script.

        This function just adds the required record header and delegates all
        work to the subclass implemention of encode_contents().

        Returns:
            bytearary: The binary version of the record that could be parsed via
                a call to UpdateRecord.FromBinary()
        """

        contents = self.encode_contents()
        record_type = self.MatchType()

        header = struct.pack("<LB3x", len(contents) + UpdateRecord.HEADER_LENGTH, record_type)

        return bytearray(header) + contents

    def encode_contents(self):
        """Encode the contents of this update record without including a record header.

        Returns:
            bytearary: The encoded contents.
        """

        raise NotImplementedError()

    @classmethod
    def LoadPlugins(cls):
        """Load all registered iotile.update_record plugins."""

        if cls.PLUGINS_LOADED:
            return

        import pkg_resources

        for entry in pkg_resources.iter_entry_points('iotile.update_record'):
            record = entry.load()
            cls.RegisterRecordType(record)

        cls.PLUGINS_LOADED = True

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this function should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): If we are asked for a match on multiple
                records in a row, because we previously returned MatchQuality.DeferMatch
                then this will indicate how many binary records are included together.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.  The only
                exception to the 0 to 100 rules is if you retern MatchQuality.DeferMatch
                which means that we are matching a multi-record statement with a single
                logical UpdateRecord.
        """

        raise NotImplementedError()

    @classmethod
    def RegisterRecordType(cls, record_class):
        """Register a known record type in KNOWN_CLASSES.

        Args:
            record_class (UpdateRecord): An update record subclass.
        """

        record_type = record_class.MatchType()
        if record_type not in UpdateRecord.KNOWN_CLASSES:
            UpdateRecord.KNOWN_CLASSES[record_type] = []

        UpdateRecord.KNOWN_CLASSES[record_type].append(record_class)

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This should be called with a binary record blob (including the record
        type header) and it will return the best record class match that it
        can find for that record.

        Args:
            record_data (bytearray): The raw record data that we wish to parse
                into an UpdateRecord subclass including its 4 byte record header.
            record_count (int): If we are asked to create a record from multiple
                records, the record_data will be passed to the record subclass
                with headers intact since there will be more than one header.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.
            DataError: If there is no matching record type registered.

        Returns:
            UpdateRecord: A subclass of UpdateRecord based on what record
                type matches the best.
        """

        # Make sure any external record types are registered
        cls.LoadPlugins()

        if len(record_data) < UpdateRecord.HEADER_LENGTH:
            raise ArgumentError("Record data is too short to contain a record header", length=len(record_data), header_length=UpdateRecord.HEADER_LENGTH)

        total_length, record_type = struct.unpack_from("<LB3x", record_data)

        if record_count == 1 and len(record_data) != total_length:
            raise ArgumentError("Record data is corrupt, embedded length does not agree with actual length", length=len(record_data), embedded_length=total_length)

        record_classes = UpdateRecord.KNOWN_CLASSES.get(record_type, [])
        if len(record_classes) == 0:
            raise DataError("No matching record type found for record", record_type=record_type, known_types=[x for x in UpdateRecord.KNOWN_CLASSES])

        best_match = MatchQuality.NoMatch
        matching_class = None

        for record_class in record_classes:
            match_data = record_data[UpdateRecord.HEADER_LENGTH:]

            if record_count > 1:
                match_data = record_data

            quality = record_class.MatchQuality(match_data, record_count)

            if quality > best_match:
                best_match = quality
                matching_class = record_class

        if best_match == MatchQuality.DeferMatch:
            raise DeferMatching(matching_class)
        elif best_match == MatchQuality.PartialMatch:
            raise DeferMatching(matching_class, matching_class.FromBinary(match_data, record_count))

        if matching_class is None:
            raise DataError("Record type found but no specific class reported a match", record_type=record_type, considered_classes=record_classes)

        return matching_class.FromBinary(match_data, record_count)

    def __ne__(self, other):
        return not self == other



from typedargs.annotate import context, docannotate
from .app import IOTileApp


@context("InfoApp")
class InfoApp(IOTileApp):
    """A Basic IOTile App that can print information about the connected device."""

    @classmethod
    def AppName(cls):
        """A unqiue name for this app so that it can be loaded by name.

        Returns:
            str: The unique name for this app module.
        """

        return 'device_info'



from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.hw.reports.signed_list_format import SignedListReport
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *
import unittest
import pytest
import os.path
import os

@pytest.fixture
def conf_hex_tracing():
    conf_file = os.path.join(os.path.dirname(__file__), 'tracing_hex_config.json')

    if '@' in conf_file or ',' in conf_file or ';' in conf_file:
        pytest.skip('Cannot pass device config because path has [@,;] in it')

    hw = HardwareManager('virtual:tracing_test@%s' % conf_file)
    hw.connect_direct('1')
    yield hw

    hw.disconnect()

@pytest.fixture
def conf_ascii_tracing():
    conf_file = os.path.join(os.path.dirname(__file__), 'tracing_ascii_config.json')

    if '@' in conf_file or ',' in conf_file or ';' in conf_file:
        pytest.skip('Cannot pass device config because path has [@,;] in it')

    hw = HardwareManager('virtual:tracing_test@%s' % conf_file)
    hw.connect_direct('1')
    yield hw

    hw.disconnect()

def test_hex_tracing(conf_hex_tracing):
    """Make sure that we can stream binary data that is not ascii printable
    """

    hw = conf_hex_tracing

    assert hw.dump_trace('raw') == ""
    assert hw.dump_trace('hex') == ""

    hw.enable_tracing()

    assert hw.dump_trace('hex') == 'abcdef1234567890abcdef1234567890abcdef1234567890abcdef1234567890'

def test_ascii_tracing(conf_ascii_tracing):
    """Make sure we can stream ascii data that is not encoded in any way
    """

    hw = conf_ascii_tracing

    assert hw.dump_trace('raw') == ""
    assert hw.dump_trace('hex') == ""

    hw.enable_tracing()

    assert hw.dump_trace('raw') == b'hello this is an acsii data stream that is somewhat long'



"""Base classes for creating and interacting with virtual IOTile devices."""

from .virtualtile import VirtualTile
from .virtualdevice import VirtualIOTileDevice
from .common_types import tile_rpc, RPCDispatcher, RPCNotFoundError, RPCInvalidArgumentsError, RPCInvalidReturnValueError, TileNotFoundError

__all__ = ['VirtualTile', 'VirtualIOTileDevice', 'tile_rpc', 'RPCDispatcher', 'TileNotFoundError', 'RPCNotFoundError', 'RPCInvalidArgumentsError', 'RPCInvalidReturnValueError']



"""Glue package shim to keep imports working since typedargs was broken into new package."""

from __future__ import unicode_literals
from functools import reduce
import pkg_resources


# Recreate all old imports
from typedargs.annotate import param, returns, context, finalizer, takes_cmdline, annotated, return_type, stringable
from typedargs.typeinfo import type_system, iprint


def load_external_components(typesys):
    """Load all external typed defined by iotile plugins.

    This allows plugins to register their own types for type annotations and
    allows all registered iotile components that have associated type libraries to
    add themselves to the global type system.
    """

    # Find all of the registered IOTile components and see if we need to add any type libraries for them
    from iotile.core.dev.registry import ComponentRegistry

    reg = ComponentRegistry()
    modules = reg.list_components()

    typelibs = reduce(lambda x, y: x+y, [reg.find_component(x).type_packages() for x in modules], [])
    for lib in typelibs:
        if lib.endswith('.py'):
            lib = lib[:-3]

        typesys.load_external_types(lib)


type_system.register_type_source('iotile.type_package', 'Pip installed type packages')
type_system.register_type_source(load_external_components, 'Local development components')



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *
import unittest
import pytest
import os.path
import os

class TestHardwareManager(unittest.TestCase):
    """
    Test to make sure that the HardwareManager is working
    """

    def setUp(self):
        self.hw = HardwareManager('none')

    def tearDown(self):
        pass

    def test_unknown_module(self):
        with pytest.raises(UnknownModuleTypeError):
            self.hw._create_proxy('UnknownTileBusModule', 8)



import unittest
import pytest
import os.path
import os
from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.hw.reports.signed_list_format import SignedListReport
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *


@pytest.fixture
def simple_hw():
    hw = HardwareManager('virtual:no_app')
    yield hw
    hw.disconnect()

def test_finding_proxy(simple_hw):
    """Make sure NO APP is matched to a TileBusProxyObject
    """

    simple_hw.connect_direct(1)
    con = simple_hw.controller()

    assert con.ModuleName() == 'NO APP'
    con = simple_hw.get(8)
    assert con.ModuleName() == 'NO APP'



"""A flexible dictionary based report format suitable for msgpack and json serialization."""

import itertools
import msgpack
import datetime
from iotile.core.exceptions import DataError
from .report import IOTileReport, IOTileReading, IOTileEvent


class FlexibleDictionaryReport(IOTileReport):
    """A list of events and readings encoded as a dictionary.

    This report format is designed to be suitable for storing in any
    format that supports key/value objects like json, msgpack, yaml,
    etc.

    Args:
        rawreport (bytearray): The raw data of this report
        signed (bool): Whether this report is signed to specify who it is from
        encrypted (bool): Whether this report is encrypted
        received_time (datetime): The time in UTC when this report was received from a device.
            If not received, the time is assumed to be utcnow().
    """

    FORMAT_TAG = "v100"

    @classmethod
    def FromReadings(cls, uuid, readings, events, report_id=IOTileReading.InvalidReadingID, selector=0xFFFF, streamer=0x100, sent_timestamp=0, received_time=None):
        """Create a flexible dictionary report from a list of readings and events.

        Args:
            uuid (int): The uuid of the deviec that this report came from
            readings (list of IOTileReading): A list of IOTileReading objects containing the data in the report
            events (list of IOTileEvent): A list of the events contained in the report.
            report_id (int): The id of the report.  If not provided it defaults to IOTileReading.InvalidReadingID.
                Note that you can specify anything you want for the report id but for actual IOTile devices
                the report id will always be greater than the id of all of the readings contained in the report
                since devices generate ids sequentially.
            selector (int): The streamer selector of this report.  This can be anything but if the report came from
                a device, it would correspond with the query the device used to pick readings to go into the report.
            streamer (int): The streamer id that this reading was sent from.
            sent_timestamp (int): The device's uptime that sent this report.
            received_time(datetime): The UTC time when this report was receievd from an IOTile device.  If it is being
                created now, received_time defaults to datetime.utcnow().

        Returns:
            FlexibleDictionaryReport: A report containing the readings and events passed in.
        """

        lowest_id = IOTileReading.InvalidReadingID
        highest_id = IOTileReading.InvalidReadingID

        for item in itertools.chain(iter(readings), iter(events)):
            if item.reading_id == IOTileReading.InvalidReadingID:
                continue

            if lowest_id == IOTileReading.InvalidReadingID or item.reading_id < lowest_id:
                lowest_id = item.reading_id
            if highest_id == IOTileReading.InvalidReadingID or item.reading_id > highest_id:
                highest_id = item.reading_id

        reading_list = [x.asdict() for x in readings]
        event_list = [x.asdict() for x in events]

        report_dict = {
            "format": cls.FORMAT_TAG,
            "device": uuid,
            "streamer_index": streamer,
            "streamer_selector": selector,
            "incremental_id": report_id,
            "lowest_id": lowest_id,
            "highest_id": highest_id,
            "device_sent_timestamp": sent_timestamp,
            "events": event_list,
            "data": reading_list
        }

        encoded = msgpack.packb(report_dict, default=_encode_datetime, use_bin_type=True)
        return FlexibleDictionaryReport(encoded, signed=False, encrypted=False, received_time=received_time)

    def decode(self):
        """Decode this report from a msgpack encoded binary blob."""

        report_dict = msgpack.unpackb(self.raw_report, raw=False)

        events = [IOTileEvent.FromDict(x) for x in report_dict.get('events', [])]
        readings = [IOTileReading.FromDict(x) for x in report_dict.get('data', [])]

        if 'device' not in report_dict:
            raise DataError("Invalid encoded FlexibleDictionaryReport that did not have a device key set with the device uuid")

        self.origin = report_dict['device']
        self.report_id = report_dict.get("incremental_id", IOTileReading.InvalidReadingID)
        self.sent_timestamp = report_dict.get("device_sent_timestamp", 0)
        self.origin_streamer = report_dict.get("streamer_index")
        self.streamer_selector = report_dict.get("streamer_selector")
        self.lowest_id = report_dict.get('lowest_id')
        self.highest_id = report_dict.get('highest_id')

        return readings, events

    def asdict(self):
        """ Return this report as a dictionary """
        return msgpack.unpackb(self.raw_report)

    def serialize(self):
        """Serialize this report including the received time."""

        raise NotImplementedError("This report format (FlexibleDictionaryReport) does not support serialization")


def _encode_datetime(obj):
    """Pack a datetime into an isoformat string."""
    if isinstance(obj, datetime.datetime):
        return obj.isoformat()

    return obj



"""An auth provider that can sign and verify with user keys stored in environment variables."""

import hashlib
import hmac
import binascii
import os
from iotile.core.exceptions import NotFoundError
from .auth_provider import AuthProvider


class EnvAuthProvider(AuthProvider):
    """Authentication provider that can sign and verify using user keys

    Keys must be defined in environment variables with the naming scheme:
    USER_KEY_ABCDEFGH where ABCDEFGH is the device uuid in hex prepended with 0s
    and expressed in capital letters.  So the device with UUID 0xab would look
    for the environment variable USER_KEY_000000AB.

    The key must be a 64 character hex string that is decoded to create a 32 byte key.
    """

    @classmethod
    def _get_key(cls, device_id):
        """Attempt to get a user key from an environment variable
        """

        var_name = "USER_KEY_{0:08X}".format(device_id)

        if var_name not in os.environ:
            raise NotFoundError("No user key could be found for devices", device_id=device_id, expected_variable_name=var_name)

        key_var = os.environ[var_name]
        if len(key_var) != 64:
            raise NotFoundError("User key in variable is not the correct length, should be 64 hex characters", device_id=device_id, key_value=key_var)

        try:
            key = binascii.unhexlify(key_var)
        except ValueError:
            raise NotFoundError("User key in variable could not be decoded from hex", device_id=device_id, key_value=key_var)

        if len(key) != 32:
            raise NotFoundError("User key in variable is not the correct length, should be 64 hex characters", device_id=device_id, key_value=key_var)

        return key

    @classmethod
    def _verify_derive_key(cls, device_id, root, **kwargs):
        report_id = kwargs.get('report_id', None)
        sent_timestamp = kwargs.get('sent_timestamp', None)

        if report_id is None or sent_timestamp is None:
            raise NotFoundError('report_id or sent_timestamp not provided in EnvAuthProvider.sign_report')

        AuthProvider.VerifyRoot(root)

        if root != AuthProvider.UserKey:
            raise NotFoundError('unsupported root key in EnvAuthProvider', root_key=root)

        root_key = cls._get_key(device_id)
        report_key = AuthProvider.DeriveReportKey(root_key, report_id, sent_timestamp)

        return report_key

    def sign_report(self, device_id, root, data, **kwargs):
        """Sign a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should sign
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signature itself must always be a bytearray stored under the
                'signature' key, however additional keys may be present depending
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to sign the data.
        """

        report_key = self._verify_derive_key(device_id, root, **kwargs)

        #We sign the SHA256 hash of the message
        message_hash = hashlib.sha256(data).digest()
        hmac_calc = hmac.new(report_key, message_hash, hashlib.sha256)
        result = bytearray(hmac_calc.digest())

        return {'signature': result, 'root_key': root}

    def verify_report(self, device_id, root, data, signature, **kwargs):
        """Verify a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should verify
            signature (bytearray): The signature attached to data that we should verify
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to verify the data due to
                an error.  If the data is simply not valid, then the function returns
                normally.
        """

        report_key = self._verify_derive_key(device_id, root, **kwargs)

        message_hash = hashlib.sha256(data).digest()
        hmac_calc = hmac.new(report_key, message_hash, hashlib.sha256)
        result = bytearray(hmac_calc.digest())

        if len(signature) == 0:
            verified = False
        elif len(signature) > len(result):
            verified = False
        elif len(signature) < len(result):
            trunc_result = result[:len(signature)]
            verified = hmac.compare_digest(signature, trunc_result)
        else:
            verified = hmac.compare_digest(signature, result)

        return {'verified': verified, 'bit_length': 8*len(signature)}

    def decrypt_report(self, device_id, root, data, **kwargs):
        """Decrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should decrypt
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The decrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to decrypt the data.
        """

        report_key = self._verify_derive_key(device_id, root, **kwargs)

        try:
            from Crypto.Cipher import AES
            import Crypto.Util.Counter
        except ImportError:
            raise NotFoundError

        ctr = Crypto.Util.Counter.new(128)

        # We use AES-128 for encryption
        encryptor = AES.new(bytes(report_key[:16]), AES.MODE_CTR, counter=ctr)

        decrypted = encryptor.decrypt(bytes(data))
        return {'data': decrypted}

    def encrypt_report(self, device_id, root, data, **kwargs):
        """Encrypt a buffer of report data on behalf of a device.

        Args:
            device_id (int): The id of the device that we should encrypt for
            root (int): The root key type that should be used to generate the report
            data (bytearray): The data that we should decrypt
            **kwargs: There are additional specific keyword args that are required
                depending on the root key used.  Typically, you must specify
                - report_id (int): The report id
                - sent_timestamp (int): The sent timestamp of the report

                These two bits of information are used to construct the per report
                signing and encryption key from the specific root key type.

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to decrypt the data.
        """

        report_key = self._verify_derive_key(device_id, root, **kwargs)

        try:
            from Crypto.Cipher import AES
            import Crypto.Util.Counter
        except ImportError:
            raise NotFoundError

        # We use AES-128 for encryption
        ctr = Crypto.Util.Counter.new(128)
        encryptor = AES.new(bytes(report_key[:16]), AES.MODE_CTR, counter=ctr)

        encrypted = encryptor.encrypt(bytes(data))
        return {'data': encrypted}



"""Tests of the CommandFile class."""

from iotile.core.utilities.command_file import CommandFile, Command


def test_encoding_idempotency():
    """Make sure encode and decode work."""

    cmd = Command("hello", [1, 'arg1', ' this is, a thing'])

    enc = CommandFile.encode(cmd)
    assert enc == u'hello {1,arg1,hex:20746869732069732c2061207468696e67}'

    dec = CommandFile.decode(enc)
    assert dec == Command("hello", ['1', 'arg1', ' this is, a thing'])


def test_dump_load():
    """Make sure we can dump and load CommandFiles."""

    cmd1 = Command("hello", ['1', 'arg1', ' this is, a thing'])
    cmd2 = Command("cmd2", [])

    cmdfile = CommandFile("TestFormat", "1.0", [cmd1, cmd2])

    out = cmdfile.dump()

    cmdfile2 = CommandFile.FromString(out)

    assert cmdfile2.commands == cmdfile.commands
    assert cmdfile2.version == cmdfile.version
    assert cmdfile2.filetype == cmdfile.filetype



import pytest
import os
from iotile.core.utilities.kvstore_json import JSONKVStore
from iotile.core.utilities.kvstore_sqlite import SQLiteKVStore


@pytest.fixture(scope='function', params=['json', 'sqlite'])
def kvstore(request):
    if request.param == 'json':
        store = JSONKVStore('testkv_store.json', respect_venv=True)
    else:
        store = SQLiteKVStore('testkv_store.db', respect_venv=True)

    store.clear()
    return store

def test_json_kvstore(kvstore):
    """Test functionality of a kvstore
    """

    with pytest.raises(KeyError):
        kvstore.get('a')

    assert kvstore.try_get('a') is None

    kvstore.set('a', 'value')

    assert kvstore.get('a') == 'value'
    assert len(kvstore.get_all()) == 1

    kvstore.clear()
    assert len(kvstore.get_all()) == 0

    kvstore.set('config:a', 'value2')

    assert kvstore.get('config:a') == 'value2'



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

import itertools
from collections import namedtuple
import json
import os.path
from future.utils import viewitems, itervalues
from past.builtins import basestring
from iotile.core.exceptions import DataError, ExternalError
from .semver import SemanticVersion, SemanticVersionRange

ReleaseStep = namedtuple('ReleaseStep', ['provider', 'args'])
ReleaseInfo = namedtuple('ReleaseInfo', ['release_date', 'dependency_versions'])
TileInfo = namedtuple('TileInfo', ['module_name', 'settings', 'architectures', 'targets', 'release_data'])


class IOTile(object):
    """
    IOTile

    A python representation of an IOTile module allowing you to inspect the products
    that its build produces and include it as a dependency in another build process.
    """

    V1_FORMAT = "v1"
    V2_FORMAT = "v2"

    def __init__(self, folder):
        self.folder = folder
        self.filter_prods = False

        modfile = os.path.join(self.folder, 'module_settings.json')

        try:
            with open(modfile, "r") as infile:
                settings = json.load(infile)
        except IOError:
            raise ExternalError("Could not load module_settings.json file, make sure this directory is an IOTile component", path=self.folder)

        file_format = settings.get('file_format', IOTile.V1_FORMAT)
        if file_format == IOTile.V1_FORMAT:
            info = self._find_v1_settings(settings)
        elif file_format == IOTile.V2_FORMAT:
            info = self._find_v2_settings(settings)
        else:
            raise DataError("Unknown file format in module_settings.json", format=file_format, path=modfile)

        self._load_settings(info)

    def _find_v1_settings(self, settings):
        """Parse a v1 module_settings.json file.

        V1 is the older file format that requires a modules dictionary with a
        module_name and modules key that could in theory hold information on
        multiple modules in a single directory.
        """

        if 'module_name' in settings:
            modname = settings['module_name']
        if 'modules' not in settings or len(settings['modules']) == 0:
            raise DataError("No modules defined in module_settings.json file")
        elif len(settings['modules']) > 1:
            raise DataError("Multiple modules defined in module_settings.json file", modules=[x for x in settings['modules']])
        else:
            modname = list(settings['modules'])[0]

        if modname not in settings['modules']:
            raise DataError("Module name does not correspond with an entry in the modules directory", name=modname, modules=[x for x in settings['modules']])

        release_info = self._load_release_info(settings)
        modsettings = settings['modules'][modname]
        architectures = settings.get('architectures', {})

        target_defs = settings.get('module_targets', {})
        targets = target_defs.get(modname, [])

        return TileInfo(modname, modsettings, architectures, targets, release_info)

    @classmethod
    def _load_release_info(cls, settings):
        if settings.get('release', False) is False:
            return None

        if 'release_date' not in settings:
            raise DataError("Release mode IOTile component did not include a release date")

        import dateutil.parser
        release_date = dateutil.parser.parse(settings['release_date'])
        dependency_versions = {x: SemanticVersion.FromString(y) for x, y in viewitems(settings.get('dependency_versions', {}))}

        return ReleaseInfo(release_date, dependency_versions)

    def _find_v2_settings(self, settings):
        archs = settings.get('architectures', {})
        mod_name = settings.get('module_name')
        release_info = self._load_release_info(settings)

        targets = settings.get('targets', [])

        return TileInfo(mod_name, settings, archs, targets, release_info)

    def _load_settings(self, info):
        """Load settings for a module."""

        modname, modsettings, architectures, targets, release_info = info
        self.settings = modsettings

        #Name is converted to all lowercase to canonicalize it
        prepend = ''
        if 'domain' in modsettings:
            prepend = modsettings['domain'].lower() + '/'

        key = prepend + modname.lower()

        #Copy over some key properties that we want easy access to
        self.name = key
        self.unique_id = key.replace('/', '_')
        self.short_name = modname
        self.targets = targets

        self.full_name = "Undefined"
        if "full_name" in self.settings:
            self.full_name = self.settings['full_name']

        #FIXME: make sure this is a list
        self.authors = []
        if "authors" in self.settings:
            self.authors = self.settings['authors']

        self.version = "0.0.0"
        if "version" in self.settings:
            self.version = self.settings['version']

        self.parsed_version = SemanticVersion.FromString(self.version)

        #Load all of the build products that can be created by this IOTile
        self.products = modsettings.get('products', {})

        #Load in the release information telling us how to release this component
        release_steps = modsettings.get('release_steps', [])
        self.release_steps = []
        self.can_release = False

        for step in release_steps:
            if 'provider' not in step:
                raise DataError("Invalid release step that did not have a provider key", step=step)

            parsed_step = ReleaseStep(provider=step['provider'], args=step.get('args', {}))
            self.release_steps.append(parsed_step)

        if len(self.release_steps) > 0:
            self.can_release = True

        self.dependency_versions = {}

        #If this is a release IOTile component, check for release information
        if release_info is not None:
            self.release = True
            self.release_date = release_info.release_date
            self.output_folder = self.folder
            self.dependency_versions = release_info.dependency_versions
        else:
            self.release = False
            self.output_folder = os.path.join(self.folder, 'build', 'output')

            #If this tile is a development tile and it has been built at least one, add in a release date
            #from the last time it was built
            if os.path.exists(os.path.join(self.output_folder, 'module_settings.json')):
                release_settings = os.path.join(self.output_folder, 'module_settings.json')

                with open(release_settings, 'rb') as infile:
                    release_dict = json.load(infile)

                import dateutil.parser
                self.release_date = dateutil.parser.parse(release_dict['release_date'])
            else:
                self.release_date = None

        #Find all of the things that this module could possibly depend on
        #Dependencies include those defined in the module itself as well as
        #those defined in architectures that are present in the module_settings.json
        #file.
        self.dependencies = []

        archs_with_deps = [viewitems(y['depends']) for _x, y in viewitems(architectures) if 'depends' in y]
        if 'depends' in self.settings:
            if not isinstance(self.settings['depends'], dict):
                raise DataError("module must have a depends key that is a dictionary", found=str(self.settings['depends']))

            archs_with_deps.append(viewitems(self.settings['depends']))

        #Find all python package needed
        self.support_wheel_depends = []

        if 'python_depends' in self.settings:
            if not isinstance(self.settings['python_depends'], list):
                raise DataError("module must have a python_depends key that is a list of strings",
                                found=str(self.settings['python_depends']))

            for python_depend in self.settings['python_depends']:
                if not isinstance(python_depend, basestring):
                    raise DataError("module must have a python_depends key that is a list of strings",
                                    found=str(self.settings['python_depends']))

                self.support_wheel_depends.append(python_depend)

        #Also search through overlays to architectures that are defined in this module_settings.json file
        #and see if those overlays contain dependencies.
        for overlay_arch in itervalues(self.settings.get('overlays', {})):
            if 'depends' in overlay_arch:
                archs_with_deps.append(viewitems(overlay_arch['depends']))

        found_deps = set()
        for dep, _ in itertools.chain(*archs_with_deps):
            name, _, version = dep.partition(',')
            unique_id = name.lower().replace('/', '_')

            version = version.strip()
            if version == '':
                version = "*"

            ver_range = SemanticVersionRange.FromString(version)

            depdict = {
                'name': name,
                'unique_id': unique_id,
                'required_version': ver_range,
                'required_version_string': version
            }

            if name not in found_deps:
                self.dependencies.append(depdict)

            found_deps.add(name)

        # Store any architectures that we find in this json file for future reference
        self.architectures = architectures

        # Setup our support package information
        self.support_distribution = "iotile_support_{0}_{1}".format(self.short_name, self.parsed_version.major)
        self.support_wheel = "{0}-{1}-py2-none-any.whl".format(self.support_distribution, self.parsed_version.pep440_string())
        self.has_wheel = False

        if len(self.proxy_modules()) > 0 or len(self.proxy_plugins()) > 0 or len(self.type_packages()) > 0 or \
           len(self.app_modules()) > 0 or len(self.build_steps()) > 0:
            self.has_wheel = True

    def include_directories(self):
        """
        Return a list of all include directories that this IOTile could provide other tiles
        """

        #Only return include directories if we're returning everything or we were asked for it
        if self.filter_prods and 'include_directories' not in self.desired_prods:
            return []

        if 'include_directories' in self.products:
            if self.release:
                joined_dirs = [os.path.join(self.output_folder, 'include', *x) for x in self.products['include_directories']]
            else:
                joined_dirs = [os.path.join(self.output_folder, *x) for x in self.products['include_directories']]
            return joined_dirs

        return []

    def libraries(self):
        """
        Return a list of all libraries produced by this IOTile that could be provided to other tiles
        """

        libs = [x[0] for x in viewitems(self.products) if x[1] == 'library']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        badlibs = [x for x in libs if not x.startswith('lib')]
        if len(badlibs) > 0:
            raise DataError("A library product was listed in a module's products without the name starting with lib", bad_libraries=badlibs)

        #Remove the prepended lib from each library name
        return [x[3:] for x in libs]

    def type_packages(self):
        """
        Return a list of the python type packages that are provided by this tile
        """

        libs = [x[0] for x in viewitems(self.products) if x[1] == 'type_package']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]

        return libs

    def linker_scripts(self):
        """
        Return a list of the linker scripts that are provided by this tile
        """

        ldscripts = [x[0] for x in viewitems(self.products) if x[1] == 'linker_script']

        if self.filter_prods:
            ldscripts = [x for x in ldscripts if x in self.desired_prods]

        # Now append the whole path so that the above comparison works based on the name of the product only
        ldscripts = [os.path.join(self.output_folder, 'linker', x) for x in ldscripts]
        return ldscripts

    def proxy_modules(self):
        """
        Return a list of the python proxy modules that are provided by this tile
        """

        libs = [x[0] for x in viewitems(self.products) if x[1] == 'proxy_module']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]
        return libs

    def app_modules(self):
        """Return a list of all of the python app module that are provided by this tile."""

        libs = [x[0] for x in viewitems(self.products) if x[1] == 'app_module']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]
        return libs

    def build_steps(self):
        """Return a list of all of the python build steps that are provided by this tile."""

        libs = [x[0] for x in viewitems(self.products) if x[1] == 'build_step']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]
        return libs

    def proxy_plugins(self):
        """
        Return a list of the python proxy plugins that are provided by this tile
        """

        libs = [x[0] for x in viewitems(self.products) if x[1] == 'proxy_plugin']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]
        return libs

    def firmware_images(self):
        """
        Return a list of the python proxy plugins that are provided by this tile
        """

        libs = [x[0] for x in viewitems(self.products) if x[1] == 'firmware_image']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.output_folder, x) for x in libs]
        return libs

    def tilebus_definitions(self):
        """
        Return a list of all tilebus definitions that this IOTile could provide other tiles
        """

        #Only return include directories if we're returning everything or we were asked for it
        if self.filter_prods and 'tilebus_definitions' not in self.desired_prods:
            return []

        if 'tilebus_definitions' in self.products:
            if self.release:
                #For released tiles, all of the tilebus definitions are copies to the same directory
                joined_dirs = [os.path.join(self.output_folder, 'tilebus', os.path.basename(os.path.join(*x))) for x in self.products['tilebus_definitions']]
            else:
                joined_dirs = [os.path.join(self.folder, 'firmware', 'src', *x) for x in self.products['tilebus_definitions']]
            return joined_dirs

        return []

    def library_directories(self):
        libs = self.libraries()

        if len(libs) > 0:
            return [os.path.join(self.output_folder)]

        return []

    def filter_products(self, desired_prods):
        """
        When asked for a product that this iotile produces, filter only those on this list
        """

        self.filter_prods = True
        self.desired_prods = set(desired_prods)

    def path(self):
        return self.folder



"""An IOTileApp that can run an update script on a device."""

from __future__ import unicode_literals, absolute_import, print_function
from builtins import input
import time
import sys
import logging
from typedargs.annotate import context, docannotate
from typedargs import iprint, type_system
from iotile.core.exceptions import HardwareError
from iotile.core.utilities.console import ProgressBar
from .app import IOTileApp
from ..update import UpdateScript


@context("DeviceUpdater")
class DeviceUpdater(IOTileApp):
    """An app that can download and run a script on an IOTile device.

    Args:
        hw (HardwareManager): A HardwareManager instance connected to a
            matching device.
        app_info (tuple): The app_tag and version of the device we are
            connected to.
        os_info (tuple): The os_tag and version of the device we are
            connected to.
        device_id (int): The UUID of the device that we are connected to.
    """

    IdleStatus = 0
    WaitingForScript = 1
    ReceivingScript = 2
    ReceivedScript = 3
    ValidatedScript = 4
    ExecutingScript = 5

    def __init__(self, hw, app_info, os_info, device_id):
        super(DeviceUpdater, self).__init__(hw, app_info, os_info, device_id)

        self._logger = logging.getLogger(__name__)
        self._con = hw.get(8, basic=True)

    @classmethod
    def AppName(cls):
        """A unqiue name for this app so that it can be loaded by name.

        Returns:
            str: The unique name for this app module.
        """

        return 'device_updater'

    @classmethod
    def _prompt_yesno(cls, message):
        valid_answer = False

        while not valid_answer:
            resp = input(message)
            resp = resp.lower()

            if resp in ['y', 'n']:
                return resp == 'y'

            print("Please type y for yes or n for no.")

    @docannotate
    def load_script(self, script_path, confirm=True, no_reboot=False):
        """Load a script from a file and run it.

        This function will load a binary update script from the file given in
        script_path, download it to the device, run it and wait for it to
        finish.

        Any errors during script execution will be raised as HardwareError
        exceptions.

        Args:
            script_path (str): The path to the binary script file (.trub extension)
                that you wish to load.
            confirm (bool): Require explicit confirmation from an interactive user
                before running the script on the device.  This defaults to true,
                which means if you want to use this function in a batch setting.
                You need to pass confirm=False.
            no_reboot (bool): Do not reboot the device after running the script.
                You typically want to reboot so this defaults to False.  If you know
                what you are doing, you can set this to True to not do a sanity reboot
                after running the script.
        """

        with open(script_path, "rb") as infile:
            indata = infile.read()

        script = UpdateScript.FromBinary(indata)

        print("Loaded script with %d actions from file %s" % (len(script.records), script_path))
        print("Running script on device 0x%X with app info (%d %s) and os_info (%d %s)" % (self._device_id, self._app_tag, self._app_version, self._os_tag, self._os_version))

        if confirm:
            show_script = self._prompt_yesno("Do you want to see the script contents before programing (y/n)? ")

            if show_script:
                print("\nScript Actions:")
                for i, record in enumerate(script.records):
                    print("%02d: %s" % (i + 1, str(record)))

                print("")

            do_load = self._prompt_yesno("Are you sure you want to program this script (y/n)? ")
            if not do_load:
                print("Aborting device update due to user request")
                return

        status, _err = self._query_status()

        if status != self.IdleStatus and not confirm:
            print("ERROR: Loading script would require user confirmation to override an incomplete prior update.")
            return

        if status in (self.ReceivedScript, self.ReceivingScript, self.WaitingForScript):
            if status == self.ReceivedScript:
                print("There is a script currently loaded in the device that was never run.")
            elif status in (self.ReceivingScript, self.WaitingForScript):
                print("WARNING: The device is currently waiting for script data from a previous update.")
                print("         Make sure that no one else is currently connected to this device before proceeding.")

            clear = self._prompt_yesno("Do you want to clear the current script to program this one (y/n)? ")
            if not clear:
                print("Aborting device update because the device was not idle")
                return

            print("Clearing loaded script (this can take up to 10 seconds)")
            self._reset_script()
        elif status in (self.ValidatedScript, self.ExecutingScript):
            print("The device is currently running a script, you must wait for it to finish.")
            return

        self.run_script(script)

    def run_script(self, script, force=False, no_reboot=False):
        """Run a script on the connected IOTile device.

        The script must an UpdateScript object.  If you are looking for an
        interactive function that will load a script from a file and show
        update information to a user, you should use load_script. This is a
        low level function that is meant to be called from a larger program.

        If this run is called from an interactive session (inside the iotile
        tool) then it will show a progress bar during download and periodic
        updates during script execution, otherwise there is no feedback until
        the script is finished.

        Args:
            script (UpdateScript): The script that we wish to run on our attached
                device.
            force (bool): If there is already a script loaded but not yet executed,
                clear that script before proceeding.
            no_reboot (bool): Do not reboot the device after running the script.
                You typically want to reboot so this defaults to False.  If you know
                what you are doing, you can set this to True to not do a sanity reboot
                after running the script.
        """

        raw_data = script.encode()

        status, _err = self._query_status()
        if status == self.ReceivedScript and force:
            self._reset_script()
        elif status != self.IdleStatus:
            raise HardwareError("Cannot run script on a remote_bridge that is not currently idle", status=status)

        self._begin_script()

        progress = ProgressBar("Downloading script", 100)

        progress.start()
        try:
            self.push_script(raw_data, progress)
        finally:
            progress.end()

        self._end_script()
        self._wait_script()

        if not no_reboot:
            iprint("Rebooting device")
            self._reboot()

    def _begin_script(self):
        """Indicate that we are going to start loading a script."""

        err, = self._con.rpc(0x21, 0x00, result_format="L", timeout=10.0)
        return err

    def _end_script(self):
        """Indicate that we are done loading a script."""

        err, = self._con.rpc(0x21, 0x02, result_format="L")
        if err != 0:
            raise HardwareError("Error ending script", error_code=err)

    def _trigger_script(self):
        """Trigger the execution of the currently loaded script."""

        err, = self._con.rpc(0x21, 0x03, result_format="L")
        if err != 0:
            raise HardwareError("Error triggering script", error_code=err)

    def _query_status(self):
        """Query the status of script loading or execution."""

        status, error = self._con.rpc(0x21, 0x04, result_format="LL")
        return status, error

    def _reset_script(self):
        err, = self._con.rpc(0x21, 0x05, result_format="L", timeout=15.0)
        if err != 0:
            raise HardwareError("Error resetting script", error_code=err)

    def _reboot(self):
        """Reboot the device."""

        self._con.reset()

    def _wait_script(self):
        """Trigger a script and then synchronously wait for it to finish processing."""

        self._trigger_script()
        status, error = self._query_status()
        if error != 0:
            raise HardwareError("Error executing remote script", error_code=error)

        error_count = 0

        iprint("Waiting for script to validate")
        while status == 3:
            time.sleep(0.1)

            try:
                status, error = self._query_status()

                if error != 0:
                    raise HardwareError("Error executing remote script", error_code=error)
            except HardwareError:
                error_count += 1
                if error_count > 2:
                    raise HardwareError("Too many errors waiting for script to finish execution", error=str(HardwareError))

            if error != 0:
                raise HardwareError("Error executing remote script", error_code=error)

        iprint("Waiting for script to finish executing")
        while status != 0:
            if type_system.interactive:
                sys.stdout.write('.')
                sys.stdout.flush()

            # Poll until the script has finished executing.  Some scripts
            # will cause the device to reset, so if that happens, make sure
            # we don't flag that as an error and cleanly reconnect.
            time.sleep(0.1)
            try:
                status, error = self._query_status()
            except HardwareError:
                error_count += 1
                if error_count > 10:
                    if type_system.interactive:
                        sys.stdout.write('\n')

                    raise HardwareError("Too many errors waiting for script to finish execution", error=str(HardwareError))

            if error != 0:
                if type_system.interactive:
                    sys.stdout.write('\n')
                raise HardwareError("Error executing remote script", error_code=error)

        if type_system.interactive:
            sys.stdout.write('\n')

    def push_script(self, data, progress=None):
        """Push a byte array into the controller as a remote script."""

        def _update_progress(current, total):
            if progress is not None:
                progress.progress(current*100/total)

        self._con.stream.send_highspeed(data, _update_progress)



"""Tests to ensure that IOTileApp matching works correctly."""

import os
import pytest
from iotile.core.hw import HardwareManager, IOTileApp
from iotile.core.dev.semver import SemanticVersionRange
from iotile.core.exceptions import HardwareError


class FakeApp(IOTileApp):
    """A basic fake app for testing."""

    @classmethod
    def MatchInfo(cls):
        return [(2100, SemanticVersionRange.FromString("^2.3.0"), 50)]

    @classmethod
    def AppName(cls):
        return 'fakeapp'

@pytest.fixture(scope="function")
def virtual_app():
    path = os.path.join(os.path.dirname(__file__), 'virtual_app_device.py')
    hw = HardwareManager(port="virtual:%s" % path)

    hw.connect(1)
    yield hw
    hw.disconnect()


@pytest.fixture(scope="function")
def register_apps():
    HardwareManager.ClearDevelopmentApps()
    HardwareManager.RegisterDevelopmentApp(FakeApp)

    yield
    HardwareManager.ClearDevelopmentApps()


def test_noapp_matching(virtual_app):
    """Make sure we throw an exception when no matching app is found."""

    hw = virtual_app

    with pytest.raises(HardwareError):
        hw.app()

    # Make sure we can still force a match
    info = hw.app('device_info')


def test_app_matching(virtual_app, register_apps):
    """Make sure matching by name and version works."""

    hw = virtual_app

    fake = hw.app()
    assert isinstance(fake, FakeApp)

    # Make sure we can still force a match
    info = hw.app('device_info')



"""The base class for all reports that are broadcast from a device without a connection.

The distinguishing feature between BroadcastReport objects and others like
IndividualReport are that BroadcastReports are not confined to a specific
connection so they needed to be treated specially.

All reports that could be Broadcast must inherit from BroadcastReport so that
all parts of CoreTools know not to drop the report if there is not an open
connection for the device in question.

Examples of ways you could receive a broadcast report are:
- bluetooth low energy broadcast readings
- MQTT posts in a topic
"""

import struct
from collections import namedtuple
import datetime
from iotile.core.exceptions import DataError
from .report import IOTileReading, IOTileReport

BroadcastHeader = namedtuple('BroadcastHeader', ['auth_type', 'reading_length', 'uuid', 'sent_timestamp', 'reserved'])


class BroadcastReport(IOTileReport):
    """Base class for all reports that are broadcast without a connection.

    This generic class supports a fixed encoding but does not yet support
    encryption/decryption or signing.  The encoding of this report is designed
    to accomodate a range of use cases and not be a binary match for any
    specific wire format that a BroadcastReport might be created from, such as
    part of a bluetooth advertisement packet.

    The encoding is designed to accomodate 1 or more IOTileReadings with a fixed
    size header and a variable length authentication block after the readings.
    """

    NO_AUTH = 0

    _HEADER_LENGTH = 16
    _AUTH_BLOCK_LENGTHS = {
        NO_AUTH: 0
    }

    ReportType = 3

    def __init__(self, rawreport, **kwargs):
        super(BroadcastReport, self).__init__(rawreport, signed=False, encrypted=False)

    @classmethod
    def _parse_header(cls, header):
        assert len(header) == cls._HEADER_LENGTH

        parsed = struct.unpack("<xBHLLL", header)
        return BroadcastHeader(*parsed)

    @classmethod
    def HeaderLength(cls):
        """Return the length of a header needed to calculate this report's length

        Returns:
            int: the length of the needed report header
        """

        return cls._HEADER_LENGTH

    @classmethod
    def ReportLength(cls, header):
        """Given a header of HeaderLength bytes, calculate the size of this report.

        Returns:
            int: The total length of the report including the header that we are passed.
        """

        parsed_header = cls._parse_header(header)

        auth_size = cls._AUTH_BLOCK_LENGTHS.get(parsed_header.auth_type)
        if auth_size is None:
            raise DataError("Unknown auth block size in BroadcastReport")

        return cls._HEADER_LENGTH + parsed_header.reading_length + auth_size

    @classmethod
    def FromReadings(cls, uuid, readings, sent_timestamp=0):
        """Generate a broadcast report from a list of readings and a uuid."""

        header = struct.pack("<BBHLLL", cls.ReportType, 0, len(readings)*16, uuid, sent_timestamp, 0)

        packed_readings = bytearray()

        for reading in readings:
            packed_reading = struct.pack("<HHLLL", reading.stream, 0, reading.reading_id, reading.raw_time, reading.value)
            packed_readings += bytearray(packed_reading)

        return BroadcastReport(bytearray(header) + packed_readings)

    def decode(self):
        """Decode this report into a list of visible readings."""

        parsed_header = self._parse_header(self.raw_report[:self._HEADER_LENGTH])

        auth_size = self._AUTH_BLOCK_LENGTHS.get(parsed_header.auth_type)
        assert auth_size is not None
        assert parsed_header.reading_length % 16 == 0

        time_base = self.received_time - datetime.timedelta(seconds=parsed_header.sent_timestamp)

        readings = self.raw_report[self._HEADER_LENGTH:self._HEADER_LENGTH + parsed_header.reading_length]
        parsed_readings = []

        for i in range(0, len(readings), 16):
            reading = readings[i:i+16]
            stream, _, reading_id, timestamp, value = struct.unpack("<HHLLL", reading)

            parsed = IOTileReading(timestamp, stream, value, time_base=time_base, reading_id=reading_id)
            parsed_readings.append(parsed)

        self.sent_timestamp = parsed_header.sent_timestamp
        self.origin = parsed_header.uuid

        return parsed_readings, []



from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.exceptions import ArgumentError
import pytest
import os.path
import os


def test_find_virtual_script_nojson():
    """Make sure we can find a virtual device script that's not importable
    """

    path = os.path.join(os.path.dirname(__file__), 'virtual_device.py')
    hw = HardwareManager(port="virtual:%s" % path)


def test_find_virtual_script_withjson():
    """Make sure we can find a script with a json
    """

    path = os.path.join(os.path.dirname(__file__), 'virtual_device.py')
    jsonpath = os.path.join(os.path.dirname(__file__), 'report_test_config_signed.json')

    hw = HardwareManager(port="virtual:%s@%s" % (path, jsonpath))


def test_invalid_virtual_script():
    path = os.path.join(os.path.dirname(__file__), 'unknown_virtual_device.py')

    with pytest.raises(ArgumentError):
        hw = HardwareManager(port="virtual:%s" % path)


def test_find_virtual_script_unknownjson():
    """Make sure we can find a script with a json
    """

    path = os.path.join(os.path.dirname(__file__), 'virtual_device.py')
    jsonpath = os.path.join(os.path.dirname(__file__), 'unknown_json_config.json')

    with pytest.raises(ArgumentError):
        hw = HardwareManager(port="virtual:%s@%s" % (path, jsonpath))



from future.utils import itervalues, viewitems
import pkg_resources
import json
import os
import traceback
import imp
import inspect
import time
from .adapter import DeviceAdapter
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.virtual.virtualdevice import RPCInvalidIDError, TileNotFoundError, RPCNotFoundError, VirtualIOTileDevice
from iotile.core.hw.reports import BroadcastReport


class VirtualAdapterAsyncChannel(object):
    """A channel for tracing and streaming data asynchronously from virtual devices
    """

    def __init__(self, adapter, iotile_id):
        self.adapter = adapter
        self.iotile_id = iotile_id

    def stream(self, report):
        """Queue data for streaming

        Args:
            report (IOTileReport): A report object to stream to a client
        """

        conn_id = self._find_connection(self.iotile_id)

        if conn_id is not None or isinstance(report, BroadcastReport):
            self.adapter._trigger_callback('on_report', conn_id, report)

    def trace(self, data):
        """Queue data for tracing

        Args:
            data (bytearray, string): Unstructured data to trace to any
                connected client.
        """

        conn_id = self._find_connection(self.iotile_id)

        if conn_id is not None:
            self.adapter._trigger_callback('on_trace', conn_id, data)

    def _find_connection(self, iotile_id):
        """Find the connection corresponding to an iotile_id
        """

        for conn_id, dev in viewitems(self.adapter.connections):
            if dev.iotile_id == iotile_id:
                return conn_id

        return None


class VirtualDeviceAdapter(DeviceAdapter):
    """Callback based adapter that gives access to one or more virtual devices

    The adapter is created and serves access to the virtual_devices that are
    found by name in the entry_point group iotile.virtual_device.

    There are two ways to passing configuration dictionaries down into a
    VirtualDevice attached to this VirtualDeviceAdapter.  Both are performed
    by encoding the port string with {{device_name}}@{{config_info}}.  You
    can either pass a path to a json file in {{config_info}} which is loaded
    as a dictionary or you can prepend a '#' to the front of config_info and
    pass a base64 encoded json string that is decoded into a dict.  Both are
    identically passed to the underlying __init__ function of the VirtualDevice

    If you choose to use a json file, the format must be::

        {
            "device":
            {
                ARGS GO HERE
            }
        }

    The required device key is to allow for this file to host information about
    multiple other kinds of iotile entities without collisions.

    If you are programmatically specifying the dictionary then it should just
    be a normal json with no parent "device" key such as:

        {
            ARGS GO HERE
        }

    Args:
        port (string): A port description that should be in the form of
            device_name1@<optional_config_json1;device_name2@optional_config_json2
    """

    # Make devices expire after a long time only
    ExpirationTime = 600000

    def __init__(self, port):
        super(VirtualDeviceAdapter, self).__init__()

        devs = port.split(';')
        loaded_devs = {}

        # This needs to be initialized before any VirtualAdapterAsyncChannels are
        # created because those could reference it
        self.connections = {}

        for dev in devs:
            name, _sep, config = dev.partition('@')

            if len(config) == 0:
                config = None

            loaded_dev = self._load_device(name, config)
            loaded_dev.start(VirtualAdapterAsyncChannel(self, loaded_dev.iotile_id))
            loaded_devs[loaded_dev.iotile_id] = loaded_dev

        self.devices = loaded_devs
        self.scan_interval = self.get_config('scan_interval', 1.0)
        self.last_scan = None

        self.set_config('probe_required', True)
        self.set_config('probe_supported', True)

    def _find_device_script(self, script_path):
        """Import a virtual device from a file rather than an installed module

        script_path must point to a python file ending in .py that contains exactly one
        VirtualIOTileDevice class definitions.  That class is loaded and executed as if it
        were installed.

        Args:
            script_path (string): The path to the script to load

        Returns:
            VirtualIOTileDevice: A subclass of VirtualIOTileDevice that was loaded from script_path
        """

        search_dir, filename = os.path.split(script_path)
        if search_dir == '':
            search_dir = './'

        if filename == '' or not os.path.exists(script_path):
            raise ArgumentError("Could not find script to load virtual device", path=script_path)

        module_name, ext = os.path.splitext(filename)
        if ext != '.py':
            raise ArgumentError("Script did not end with .py", filename=filename)

        try:
            file = None
            file, pathname, desc = imp.find_module(module_name, [search_dir])
            mod = imp.load_module(module_name, file, pathname, desc)
        finally:
            if file is not None:
                file.close()

        devs = [x for x in itervalues(mod.__dict__) if inspect.isclass(x) and issubclass(x, VirtualIOTileDevice) and x != VirtualIOTileDevice]
        if len(devs) == 0:
            raise ArgumentError("No VirtualIOTileDevice subclasses were defined in script", path=script_path)
        elif len(devs) > 1:
            raise ArgumentError("More than one VirtualIOTileDevice subclass was defined in script", path=script_path, devices=devs)

        return devs[0]

    def _load_device(self, name, config):
        """Load a device either from a script or from an installed module
        """

        if config is None:
            config_dict = {}
        elif isinstance(config, dict):
            config_dict = config
        elif config[0] == '#':
            # Allow passing base64 encoded json directly in the port string to ease testing.
            import base64
            config_str = base64.b64decode(config[1:])
            config_dict = json.loads(config_str)
        else:
            try:
                with open(config, "r") as conf:
                    data = json.load(conf)
            except IOError as exc:
                raise ArgumentError("Could not open config file", error=str(exc), path=config)

            if 'device' not in data:
                raise ArgumentError("Invalid configuration file passed to VirtualDeviceAdapter", device_name=name, config_path=config, missing_key='device')

            config_dict = data['device']

        if name.endswith('.py'):
            device_factory = self._find_device_script(name)
            return device_factory(config_dict)

        seen_names = []
        for entry in pkg_resources.iter_entry_points('iotile.virtual_device'):
            if entry.name == name:
                device_factory = entry.load()
                return device_factory(config_dict)

            seen_names.append(entry.name)

        raise ArgumentError("Could not find virtual_device by name", name=name, known_names=seen_names)

    def can_connect(self):
        """Can this adapter have another simultaneous connection

        There is no limit on the number of simultaneous virtual connections

        Returns:
            bool: whether another connection is allowed
        """

        return True

    def connect_async(self, connection_id, connection_string, callback):
        """Asynchronously connect to a device

        Args:
            connection_id (int): A unique identifer that will refer to this connection
            connection_string (string): A DeviceAdapter specific string that can be used to connect to
                a device using this DeviceAdapter.
            callback (callable): A function that will be called when the connection attempt finishes as
                callback(conection_id, adapter_id, success: bool, failure_reason: string or None)
        """

        id_number = int(connection_string)
        if id_number not in self.devices:
            if callback is not None:
                callback(connection_id, self.id, False, "could not find device to connect to")
            return

        if id_number in [x.iotile_id for x in itervalues(self.connections)]:
            if callback is not None:
                callback(connection_id, self.id, False, "device was already connected to")
            return

        dev = self.devices[id_number]
        dev.connected = True

        self.connections[connection_id] = dev

        if callback is not None:
            callback(connection_id, self.id, True, "")

    def disconnect_async(self, conn_id, callback):
        """Asynchronously disconnect from a connected device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.connected = False
        del self.connections[conn_id]

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def _open_rpc_interface(self, conn_id, callback):
        """Open the RPC interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.open_rpc_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def _open_streaming_interface(self, conn_id, callback):
        """Open the streaming interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        reports = dev.open_streaming_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

        for report in reports:
            self._trigger_callback('on_report', conn_id, report)

    def _open_tracing_interface(self, conn_id, callback):
        """Open the tracing interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")

            return

        dev = self.connections[conn_id]
        traces = dev.open_tracing_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

        for trace in traces:
            self._trigger_callback('on_trace', conn_id, trace)

    def _open_script_interface(self, conn_id, callback):
        """Open the script interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.open_script_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def send_rpc_async(self, conn_id, address, rpc_id, payload, timeout, callback):
        """Asynchronously send an RPC to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            address (int): the addres of the tile that we wish to send the RPC to
            rpc_id (int): the 16-bit id of the RPC we want to call
            payload (bytearray): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute
            callback (callable): A callback for when we have finished the RPC.  The callback will be called as"
                callback(connection_id, adapter_id, success, failure_reason, status, payload)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
                'status': the one byte status code returned for the RPC if success == True else None
                'payload': a bytearray with the payload returned by RPC if success == True else None
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, 'Device is not in connected state', None, None)
            return

        dev = self.connections[conn_id]

        status = (1 << 6)
        try:
            response = dev.call_rpc(address, rpc_id, bytes(payload))
            if len(response) > 0:
                status |= (1 << 7)
        except (RPCInvalidIDError, RPCNotFoundError):
            status = 2
            response = bytes()
        except TileNotFoundError:
            status = 0xFF
            response = bytes()
        except Exception:
            #Don't allow exceptions or we will deadlock
            status = 3
            response = bytes()

            print("*** EXCEPTION OCCURRED IN RPC ***")
            traceback.print_exc()
            print("*** END EXCEPTION ***")

        response = bytearray(response)
        callback(conn_id, self.id, True, "", status, response)

    def send_script_async(self, conn_id, data, progress_callback, callback):
        """Asynchronously send a a script to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            data (bytes or bytearray): the script to send to the device
            progress_callback (callable): A function to be called with status on our progress, called as:
                progress_callback(done_count, total_count)
            callback (callable): A callback for when we have finished sending the script.  The callback will be called as"
                callback(connection_id, adapter_id, success, failure_reason)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, 'Device is not in connected state')
            return

        # Simulate some progress callbacks (0, 50%, 100%)
        progress_callback(0, len(data))
        dev = self.connections[conn_id]
        dev.script = data
        progress_callback(len(data)//2, len(data))
        progress_callback(len(data), len(data))

        callback(conn_id, self.id, True, None)

    def _send_scan_event(self, device):
        """Send a scan event from a device
        """

        info = {
            'connection_string': str(device.iotile_id),
            'uuid': device.iotile_id,
            'signal_strength': 100
        }
        self._trigger_callback('on_scan', self.id, info, self.ExpirationTime)

    def periodic_callback(self):
        if self.last_scan is None or time.time() > (self.scan_interval + self.last_scan):
            for dev in itervalues(self.devices):
                self._send_scan_event(dev)

    def stop_sync(self):
        for dev in itervalues(self.devices):
            dev.stop()

    def probe_async(self, callback):
        """Send advertisements for all connected virtual devices.

        Args:
            callback (callable): A callback for when the probe operation has completed.
                callback should have signature callback(adapter_id, success, failure_reason) where:
                    success: bool
                    failure_reason: None if success is True, otherwise a reason for why we could not probe
        """

        for dev in itervalues(self.devices):
                self._send_scan_event(dev)

        callback(self.id, True, None)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)
from builtins import str, input
from future.utils import viewitems
import sys
import os
import threading
import traceback
import argparse
import logging

from typedargs.shell import HierarchicalShell
from typedargs import type_system
from iotile.core.exceptions import IOTileException, APIError
from iotile.core.dev.registry import ComponentRegistry
import iotile.core.hw.transport.cmdstream as cmdstream

DESCRIPTION= \
"""Create an interactive shell that explores the IOTile API.

This tool allows you to run commands that are defined in either CoreTools, or
in a registered plugin.  You can do things like build IOTile firmware or
control an IOTile device.

If you wish enable logging to debug something that is not working correctly,
you can do so by passing a combination of -v, -l, -e and -i flags as needed.
See iotile --help for more details.

**NB, if you want to pass global arguments to enable logging you must do so
before the first command you pass otherwise the global arguments will be
interpreted as arguments to your commands.**
"""

def timeout_thread_handler(timeout, stop_event):
    """A background thread to kill the process if it takes too long.

    Args:
        timeout (float): The number of seconds to wait before killing
            the process.
        stop_event (Event): An optional event to cleanly stop the background
            thread if required during testing.
    """

    stop_happened = stop_event.wait(timeout)
    if stop_happened is False:
        print("Killing program due to %f second timeout" % timeout)

    os._exit(2)


def create_parser():
    """Create the argument parser for iotile."""
    parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-v', '--verbose', action="count", default=0, help="Increase logging level (goes error, warn, info, debug)")
    parser.add_argument('-l', '--logfile', help="The file where we should log all logging messages")
    parser.add_argument('-i', '--include', action="append", default=[], help="Only include the specified loggers")
    parser.add_argument('-e', '--exclude', action="append", default=[], help="Exclude the specified loggers, including all others")
    parser.add_argument('-q', '--quit', action="store_true", help="Do not spawn a shell after executing any commands")
    parser.add_argument('-t', '--timeout', type=float, help="Do not allow this process to run for more than a specified number of seconds.")
    parser.add_argument('commands', nargs=argparse.REMAINDER, help="The command(s) to execute")

    return parser


def parse_global_args(argv):
    """Parse all global iotile tool arguments.

    Any flag based argument at the start of the command line is considered as
    a global flag and parsed.  The first non flag argument starts the commands
    that are passed to the underlying hierarchical shell.

    Args:
        argv (list): The command line for this command

    Returns:
        Namespace: The parsed arguments, with all of the commands that should
            be executed in an iotile shell as the attribute 'commands'
    """

    parser = create_parser()
    args = parser.parse_args(argv)

    should_log = args.include or args.exclude or (args.verbose > 0)
    verbosity = args.verbose

    root = logging.getLogger()

    if should_log:
        formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname).3s %(name)s %(message)s', '%y-%m-%d %H:%M:%S')
        if args.logfile:
            handler = logging.FileHandler(args.logfile)
        else:
            handler = logging.StreamHandler()

        handler.setFormatter(formatter)

        if args.include and args.exclude:
            print("You cannot combine whitelisted (-i) and blacklisted (-e) loggers, you must use one or the other.")
            sys.exit(1)

        loglevels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
        if verbosity >= len(loglevels):
            verbosity = len(loglevels) - 1

        level = loglevels[verbosity]

        if args.include:
            for name in args.include:
                logger = logging.getLogger(name)
                logger.setLevel(level)
                logger.addHandler(handler)

            root.addHandler(logging.NullHandler())
        else:
            # Disable propagation of log events from disabled loggers
            for name in args.exclude:
                logger = logging.getLogger(name)
                logger.disabled = True

            root.setLevel(level)
            root.addHandler(handler)
    else:
        root.addHandler(logging.NullHandler())

    return args


def setup_completion(shell):
    """Setup readline to tab complete in a cross platform way."""

    #Handle special case of importing pyreadline on Windows
    #See: http://stackoverflow.com/questions/6024952/readline-functionality-on-windows-with-python-2-7
    import glob
    try:
        import readline
    except ImportError:
        import pyreadline as readline

    def _complete(text, state):
        buf = readline.get_line_buffer()
        if buf.startswith('help ') or " " not in buf:
            return [x for x in shell.valid_identifiers() if x.startswith(text)][state]

        return (glob.glob(os.path.expanduser(text)+'*')+[None])[state]

    readline.set_completer_delims(' \t\n;')
    #Handle Mac OS X special libedit based readline
    #See: http://stackoverflow.com/questions/7116038/python-tab-completion-mac-osx-10-7-lion
    if readline.__doc__ is not None and 'libedit' in readline.__doc__:
        readline.parse_and_bind("bind ^I rl_complete")
    else:
        readline.parse_and_bind("tab: complete")

    readline.set_completer(_complete)


def main(argv=None):
    """Run the iotile shell tool.

    You can optionally pass the arguments that should be run
    in the argv parameter.  If nothing is passed, the args
    are pulled from sys.argv.

    The return value of this function is the return value
    of the shell command.
    """

    if argv is None:
        argv = sys.argv[1:]

    args = parse_global_args(argv)

    type_system.interactive = True
    line = args.commands

    timeout_thread = None
    timeout_stop_event = threading.Event()

    if args.timeout:
        timeout_thread = threading.Thread(target=timeout_thread_handler, args=(args.timeout, timeout_stop_event))
        timeout_thread.daemon = True
        timeout_thread.start()

    shell = HierarchicalShell('iotile')

    shell.root_add("registry", "iotile.core.dev.annotated_registry,registry")
    shell.root_add("config", "iotile.core.dev.config,ConfigManager")
    shell.root_add('hw', "iotile.core.hw.hwmanager,HardwareManager")

    reg = ComponentRegistry()
    plugins = reg.list_plugins()
    for key, val in viewitems(plugins):
        shell.root_add(key, val)

    finished = False

    try:
        if len(line) > 0:
            finished = shell.invoke(line)
    except IOTileException as exc:
        print(exc.format())
        #if the command passed on the command line fails, then we should
        #just exit rather than drop the user into a shell.
        cmdstream.do_final_close()
        return 1
    except Exception:  # pylint:disable=broad-except; We need to make sure we always call cmdstream.do_final_close()
        #Catch all exceptions because otherwise we won't properly close cmdstreams
        #since the program will be said to except 'abnormally'
        traceback.print_exc()
        cmdstream.do_final_close()
        return 1

    # If the user tells us to never spawn a shell, make sure we don't
    # Also, if we finished our command and there is no context, quit now
    if args.quit or finished:
        cmdstream.do_final_close()
        return 0

    setup_completion(shell)

    #We ended the initial command with a context, start a shell
    try:
        while True:
            try:
                linebuf = input("(%s) " % shell.context_name())

                # Skip comments automatically
                if len(linebuf) > 0 and linebuf[0] == '#':
                    continue
            except KeyboardInterrupt:
                print("")
                continue

            #Catch exception outside the loop so we stop invoking submethods if a parent
            #fails because then the context and results would be unpredictable
            try:
                finished = shell.invoke_string(linebuf)
            except KeyboardInterrupt:
                print("")
                if timeout_stop_event.is_set():
                    break
            except IOTileException as exc:
                print(exc.format())
            except Exception:  #pylint:disable=broad-except; We want to make sure the iotile tool never crashes when in interactive shell mode
                traceback.print_exc()

            if shell.finished():
                break

    #Make sure to catch ^C and ^D so that we can cleanly dispose of subprocess resources if
    #there are any.
    except EOFError:
        print("")
    except KeyboardInterrupt:
        print("")
    finally:
        #Make sure we close any open CMDStream communication channels so that we don't lockup at exit
        cmdstream.do_final_close()

    # Make sure we cleanly join our timeout thread before exiting
    if timeout_thread is not None:
        timeout_stop_event.set()
        timeout_thread.join()



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

#exceptions.py

import iotile.core.exceptions


class RPCError(iotile.core.exceptions.HardwareError):
    pass

class ModuleBusyError(RPCError):
    def __init__(self, address, **kwargs):
        super(ModuleBusyError, self).__init__("Module responded that it was busy", address=address, **kwargs)

class UnsupportedCommandError(RPCError):
    def __init__(self, **kwargs):
        super(UnsupportedCommandError, self).__init__("Module did not support the specified command", **kwargs)

class ModuleNotFoundError(RPCError):
    def __init__(self, address, **kwargs):
        super(ModuleNotFoundError, self).__init__("Module was not found or did not respond", address=address, **kwargs)

class StreamNotConnectedError(RPCError):
    def __init__(self, **kwargs):
        super(StreamNotConnectedError, self).__init__("Stream was not connected to any MoMo devices", **kwargs)

class StreamOperationNotSupportedError(RPCError):
    def __init__(self, **kwargs):
        super(StreamOperationNotSupportedError, self).__init__("Underlying command stream does not support the required operation", **kwargs)

class UnknownModuleTypeError(iotile.core.exceptions.TypeSystemError):
    pass



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

from iotile.core.hw.transport.cmdstream import CMDStream
from iotile.core.hw.commands import RPCCommand
from iotile.core.exceptions import *
from iotile.core.hw.exceptions import *
import hashlib
import json

class RecordedStream (CMDStream):
	def __init__(self, filepath, conn_string, record=None):
		self.filepath = filepath
		self.recording = {}
		self.calls = {}

		if record is not None:
			raise ValidationError("You cannot record a session based on an already recorded stream")

		if self.filepath is not None:
			self._load_recording()

		super(RecordedStream, self).__init__(filepath, conn_string)

	def _connect(self, conn_string, wait=None):
		if conn_string not in self.recording:
			raise HardwareError("Attempting to connect to an unrecorded device in a RecordedStream", connection_string=conn_string, recorded_devices=self.recording.keys())

	def _send_rpc(self, address, feature, command, *args, **kwargs):
		rpc = RPCCommand(address, feature, command, *args)

		payload = rpc._format_args()
		payload = payload[:rpc.spec]

		hashval = self._hash_call(address, feature, command, payload)
		if hashval.hexdigest() not in self.recording[self.connection_string]:
			raise HardwareError("Attempting to make a nonrecorded RPC on a RecordedStream", address=address, feature=feature, command=command, payload=payload)

		#Figure out how many calls we've made to this rpc, then retrieve the response to the ith call and update our call counter
		if self.connection_string not in self.calls:
			self.calls[self.connection_string] = {}

		if hashval.hexdigest() not in self.calls[self.connection_string]:
			self.calls[self.connection_string][hashval.hexdigest()] = 0

		call_count = self.calls[self.connection_string][hashval.hexdigest()]

		if call_count >= len(self.recording[self.connection_string][hashval.hexdigest()]):
			raise HardwareError("Attempted to call a recorded RPC more times than were recorded", recorded_calls=len(self.recording[self.connection_string][hashval.hexdigest()]))

		status, payload = self.recording[self.connection_string][hashval.hexdigest()][call_count]

		self.calls[self.connection_string][hashval.hexdigest()] += 1

		return status, payload

	def _reset(self):
		pass

	def _close(self):
		pass

	def _hash_call(self, addr, feature, command, payload):
		msg = bytearray(chr(addr) + chr(feature) + chr(command) + payload)
		return hashlib.sha256(msg)

	def _load_recording(self):
		with open(self.filepath, "r") as f:
			data = json.load(f)

		for dev in data:
			res = data[dev]
			self.recording[dev] = {}

			for call, resp in res:
				addr, feature, command, payload = call.split(",")

				payload = bytearray.fromhex(payload)
				addr = int(addr, 0)
				feature = int(feature, 0)
				command = int(command, 0)

				assert addr >= 0 and addr <= 255
				assert feature >= 0 and feature <= 255
				assert command >= 0 and command <= 255

				hashval = self._hash_call(addr, feature, command, payload)

				#Process response, which is in the format hex status, hex payload
				status, payload = resp.split(",")

				status = int(status, 0)
				payload = bytearray.fromhex(payload)

				#Make sure HasData bit is always set whenever there is actually data
				if len(payload) != 0:
					status |= (1 << 7)

				#Allow multiple calls to the same function returning different things each time

				if hashval.hexdigest() not in self.recording[dev]:
					self.recording[dev][hashval.hexdigest()] = []

				self.recording[dev][hashval.hexdigest()].append((status, payload))



"""Test our update script generation and parsing."""

import pytest
from iotile.core.exceptions import ArgumentError, DataError
from iotile.core.hw.update.records import *
from iotile.core.hw import UpdateScript


def test_basic_script_parsing():
    """Make sure we can parse an update script."""

    reflash = ReflashTileRecord(1, bytearray(100), 0x1000)
    unknown = UnknownRecord(128, bytearray(15))

    script = UpdateScript([reflash, unknown])

    encoded = script.encode()

    script2 = UpdateScript.FromBinary(encoded)

    assert len(script2.records) == 2

    with pytest.raises(DataError):
        UpdateScript.FromBinary(encoded, allow_unknown=False)

    assert script == script2
    assert not script != script2

    script3 = UpdateScript([reflash, unknown, unknown])

    assert script3 != script2

def test_set_version():
    """Ensure set_version script performs as intended"""
    app_record = SetDeviceTagRecord(app_tag=12, app_version='3.4')
    os_record = SetDeviceTagRecord(os_tag=56, os_version='7.8')
    both_record = SetDeviceTagRecord(app_tag=56, app_version='7.8', os_tag=12, os_version='3.4')

    script = UpdateScript([app_record, os_record, both_record])
    encoded = script.encode()
    script2 = UpdateScript.FromBinary(encoded)

    assert script2.records[0].update_app == True
    assert script2.records[0].update_os == False
    assert script2.records[0].app_tag == 12
    assert script2.records[0].app_version == '3.4'
    assert script2.records[0].os_tag == None
    assert script2.records[0].os_version == '0.0'

    assert script2.records[1].update_app == False
    assert script2.records[1].update_os == True
    assert script2.records[1].app_tag == None
    assert script2.records[1].app_version == '0.0'
    assert script2.records[1].os_tag == 56
    assert script2.records[1].os_version == '7.8'

    assert script2.records[2].update_app == True
    assert script2.records[2].update_os == True
    assert script2.records[2].app_tag == 56
    assert script2.records[2].app_version == '7.8'
    assert script2.records[2].os_tag == 12
    assert script2.records[2].os_version == '3.4'

    assert str(script2.records[0]) == u'Set device app to (tag:12 version:3.4)'
    assert str(script2.records[1]) == u'Set device os to (tag:56, version:7.8)'
    assert str(script2.records[2]) == u'Set device os to (tag:12, version:3.4) and app to (tag:56, version:7.8)'



"""A parser/generator for a list of commands with arguments.

All commands are single words with arguments as follows:

command_name {arg1, arg2, arg3}

If the contents of argN contains a , or begins or ends with whitespace
it is utf-8 encoded and stored as hex of the form hex:xxxxx
"""

from __future__ import unicode_literals
from builtins import str
from collections import namedtuple
from binascii import hexlify, unhexlify
from iotile.core.exceptions import DataError


Command = namedtuple("Command", ["name", "args"])

class CommandFile(object):
    """A versioned and typed list of commands.

    Args:
        filetype (str): The name of the file type
        version (str): An X.Y.Z version number
        commands (list): A list of Command objects used to
            initialize the command file.
    """


    def __init__(self, filetype, version, commands=None):
        if commands is None:
            commands = []

        self.commands = commands
        self.validators = {}
        self.filetype = filetype
        self.version = version

    def add(self, command, *args):
        """Add a command to this command file.

        Args:
            command (str): The command to add
            *args (str): The parameters to call the command with
        """

        cmd = Command(command, args)
        self.commands.append(cmd)

    def save(self, outpath):
        """Save this command file as an ascii file.

        Agrs:
            outpath (str): The output path to save.
        """

        with open(outpath, "w") as outfile:
            outfile.write(self.dump())

    def dump(self):
        """Dump all commands in this object to a string.

        Returns:
            str: An encoded list of commands separated by
                \n characters suitable for saving to a file.
        """

        out = []

        out.append(self.filetype)
        out.append("Format: {}".format(self.version))
        out.append("Type: ASCII")
        out.append("")

        for cmd in self.commands:
            out.append(self.encode(cmd))

        return "\n".join(out) + "\n"

    @classmethod
    def FromString(cls, indata):
        """Load a CommandFile from a string.

        The string should be produced from a previous call to
        encode.

        Args:
            indata (str): The encoded input data.

        Returns:
            CommandFile: The decoded CommandFile object.
        """

        lines = [x.strip() for x in indata.split("\n") if not x.startswith('#') and not x.strip() == ""]

        if len(lines) < 3:
            raise DataError("Invalid CommandFile string that did not contain 3 header lines", lines=lines)

        fmt_line, version_line, ascii_line = lines[:3]

        if not version_line.startswith("Format: "):
            raise DataError("Invalid format version that did not start with 'Format: '", line=version_line)

        version = version_line[8:]

        if ascii_line != "Type: ASCII":
            raise DataError("Unknown file type line (expected Type: ASCII)", line=ascii_line)

        cmds = [cls.decode(x) for x in lines[3:]]
        return CommandFile(fmt_line, version, cmds)

    @classmethod
    def FromFile(cls, inpath):
        """Load a CommandFile from a path.

        Args:
            inpath (str): The path to the file to load

        Returns:
            CommandFile: The decoded CommandFile object.
        """

        with open(inpath, "r") as infile:
            indata = infile.read()

        return cls.FromString(indata)

    @classmethod
    def encode(cls, command):
        """Encode a command as an unambiguous string.

        Args:
            command (Command): The command to encode.

        Returns:
            str: The encoded command
        """

        args = []
        for arg in command.args:
            if not isinstance(arg, str):
                arg = str(arg)

            if "," in arg or arg.startswith(" ") or arg.endswith(" ") or arg.startswith("hex:"):
                arg = "hex:{}".format(hexlify(arg.encode('utf-8')).decode('utf-8'))

            args.append(arg)

        argstr = ""

        if len(args) > 0:
            argstr = " {" + ",".join(args) + "}"

        return command.name + argstr

    @classmethod
    def decode(cls, command_str):
        """Decode a string encoded command back into a Command object.

        Args:
            command_str (str): The encoded command string output from a
                previous call to encode.

        Returns:
            Command: The decoded Command object.
        """

        name, _, arg = command_str.partition(" ")

        args = []

        if len(arg) > 0:
            if arg[0] != '{' or arg[-1] != '}':
                raise DataError("Invalid command, argument is not contained in { and }", arg=arg, cmd=name)

            arg = arg[1:-1]
            args = arg.split(",")

        proc = []

        for arg in args:
            if arg.startswith("hex:"):
                arg = unhexlify(arg[4:]).decode('utf-8')

            proc.append(arg)

        return Command(name, proc)



# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

import json
import os.path
import sys
import os
import platform
from iotile.core.utilities.paths import settings_directory

class JSONKVStore(object):
    """A Key Value store based on flat json files with atomic write semantics

    This is intended as a drop in replacement for SQLiteKVStore.  Note that the
    implementation is not meant to be efficient in the sense of caching the file
    in memory.  Instead it has read-through and write-through semantics where the
    file is reloaded every time a request is made.

    Args:
        name (string): The name of the file to use as a persistent store for this KVStore
        folder (string): Optional folder to store the file.  If None, the system default
            settings directory is used
        respect_venv (bool): Make folder relative to the current virtual environment if there
            is one.
    """

    DefaultFolder = settings_directory()

    def __init__(self, name, folder=None, respect_venv=False):
        if folder is None:
            folder = JSONKVStore.DefaultFolder

        #If we are relative to a virtual environment, place the registry into that virtual env
        #Support both virtualenv and pythnon 3 venv
        if respect_venv and hasattr(sys, 'real_prefix'):
            folder = sys.prefix
        elif respect_venv and hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix:
            folder = sys.prefix

        if not os.path.exists(folder):
            os.makedirs(folder, 0o755)

        jsonfile = os.path.join(folder, name)

        self.file = jsonfile

    def _load_file(self):
        """Load all entries from json backing file
        """

        if not os.path.exists(self.file):
            return {}

        with open(self.file, "r") as infile:
            data = json.load(infile)

        return data

    def _save_file(self, data):
        """Attempt to atomically save file by saving and then moving into position

        The goal is to make it difficult for a crash to corrupt our data file since
        the move operation can be made atomic if needed on mission critical filesystems.
        """

        if platform.system() == 'Windows':
            with open(self.file, "w") as outfile:
                json.dump(data, outfile)
        else:
            newpath = self.file + '.new'

            with open(newpath, "w") as outfile:
                json.dump(data, outfile)

            os.rename(
                os.path.realpath(newpath),
                os.path.realpath(self.file)
            )

    def get(self, key):
        """Get a value by its key

        Args:
            key (string): The key used to store this value

        Returns:
            value (string): The value associated to the key

        Raises:
            KeyError: if the key was not found
        """

        data = self._load_file()
        return data[key]

    def get_all(self):
        """Return a list of all (key, value) tuples in the kv store

        Returns:
            list(string, string): A list of key, value pairs
        """

        data = self._load_file()
        return data.items()

    def remove(self, key):
        """Remove a key from the data store

        Args:
            key (string): The key to remove

        Raises:
            KeyError: if the key was not found
        """

        data = self._load_file()
        del data[key]
        self._save_file(data)

    def try_get(self, key):
        """Try to get a value by its key, returning None if not found

        Args:
            key (string): The key used to store this value

        Returns:
            value (string): The value associated to the key or None
        """

        data = self._load_file()
        return data.get(key, None)

    def set(self, key, value):
        """Set the value of a key

        Args:
            key (string): The key used to store this value
            value (string): The value to store
        """

        data = self._load_file()
        data[key] = value
        self._save_file(data)

    def clear(self):
        """Clear all values from this kv store
        """

        self._save_file({})



"""Canonical list of config variables defined by iotile-core."""


def get_variables():
    """Get a dictionary of configuration variables."""

    prefix = "core"

    conf_vars = []
    conf_vars.append(["default-port", "string", "The default port to use for HardwareManager sessions if none is given"])

    return prefix, conf_vars



"""Virtual device for testing app matching."""

from iotile.core.hw.virtual.virtualdevice import VirtualIOTileDevice, rpc


class AppInfoVirtualDevice(VirtualIOTileDevice):
    """Virtual device that acts like a blank tile with just an executive loaded.

    Args:
        args (dict): Any arguments that you want to pass to create this device.
            None are supported.
    """

    def __init__(self, args):
        super(AppInfoVirtualDevice, self).__init__(1, 'abcdef')

    @rpc(8, 0x0004, "", "H6sBBBB")
    def status(self):
        """Return the name of the controller as a 6 byte string
        """

        status = (1 << 1) | (1 << 0)  # Configured and running

        return [0xFFFF, self.name, 1, 0, 0, status]

    @rpc(8, 0x1008, "", "12xLL")
    def app_os_info(self):
        """Return os/app info for this device."""

        os_tag = 2000
        os_version = (1, 2)
        app_tag = 2100
        app_version = (2, 3)

        os_info = (os_version[0] << 26 | os_version[1] << 20 | os_tag)
        app_info = (app_version[0] << 26 | app_version[1] << 20 | app_tag)

        return [os_info, app_info]



from .individual_format import IndividualReadingReport
from .report import IOTileReading, IOTileReport
from .signed_list_format import SignedListReport
from .broadcast import BroadcastReport
from .parser import IOTileReportParser
from .flexible_dictionary import FlexibleDictionaryReport


__all__ = ['IndividualReadingReport', 'IOTileReport', 'IOTileReading', 'BroadcastReport', 'SignedListReport', 'FlexibleDictionaryReport', 'IOTileReportParser']



"""A list of update records that specify a script for updating a device."""

from __future__ import (print_function, absolute_import, unicode_literals)
import struct
import hashlib
import logging
from hmac import compare_digest
from binascii import hexlify
from collections import namedtuple
from iotile.core.exceptions import ArgumentError, DataError
from .record import UpdateRecord, DeferMatching
from .records import UnknownRecord

ScriptHeader = namedtuple('ScriptHeader', ['header_length', 'authenticated', 'integrity_checked', 'encrypted'])


class UpdateScript(object):
    """An update script that consists of a list of UpdateRecord objects.

    Args:
        records (list of UpdateRecord): The records that make up this script.
    """

    SCRIPT_MAGIC = 0x1F2E3D4C
    SCRIPT_HEADER_LENGTH = 24

    logger = logging.getLogger(__name__)

    def __init__(self, records):
        self.records = records

    @classmethod
    def ParseHeader(cls, script_data):
        """Parse a script integrity header.

        This function makes sure any integrity hashes are correctly parsed and
        returns a ScriptHeader structure containing the information that it
        was able to parse out.

        Args:
            script_data (bytearray): The script that we should parse.

        Raises:
            ArgumentError: If the script contains malformed data that
                cannot be parsed.

        Returns:
            ScriptHeader: The parsed script header information
        """

        if len(script_data) < UpdateScript.SCRIPT_HEADER_LENGTH:
            raise ArgumentError("Script is too short to contain a script header", length=len(script_data), header_length=UpdateScript.SCRIPT_HEADER_LENGTH)

        embedded_hash, magic, total_length = struct.unpack_from("<16sLL", script_data)
        if magic != UpdateScript.SCRIPT_MAGIC:
            raise ArgumentError("Script has invalid magic value", expected=UpdateScript.SCRIPT_MAGIC, found=magic)

        if total_length != len(script_data):
            raise ArgumentError("Script length does not match embedded length", embedded_length=total_length, length=len(script_data))

        hashed_data = script_data[16:]

        sha = hashlib.sha256()
        sha.update(hashed_data)
        hash_value = sha.digest()[:16]

        if not compare_digest(embedded_hash, hash_value):
            raise ArgumentError("Script has invalid embedded hash", embedded_hash=hexlify(embedded_hash), calculated_hash=hexlify(hash_value))

        return ScriptHeader(UpdateScript.SCRIPT_HEADER_LENGTH, False, True, False)

    @classmethod
    def FromBinary(cls, script_data, allow_unknown=True):
        """Parse a binary update script.

        Args:
            script_data (bytearray): The binary data containing the script.
            allow_unknown (bool): Allow the script to contain unknown records
                so long as they have correct headers to allow us to skip them.
        Raises:
            ArgumentError: If the script contains malformed data that cannot
                be parsed.
            DataError: If the script contains unknown records and allow_unknown=False

        Returns:
            UpdateScript: The parsed update script.
        """

        curr = 0
        records = []

        header = cls.ParseHeader(script_data)
        curr = header.header_length

        cls.logger.debug("Parsed script header: %s, skipping %d bytes", header, curr)

        record_count = 0
        record_data = bytearray()
        partial_match = None
        match_offset = 0

        while curr < len(script_data):
            if len(script_data) - curr < UpdateRecord.HEADER_LENGTH:
                raise ArgumentError("Script ended with a partial record", remaining_length=len(script_data) - curr)

            # Add another record to our current list of records that we're parsing

            total_length, record_type = struct.unpack_from("<LB", script_data[curr:])
            cls.logger.debug("Found record of type %d, length %d", record_type, total_length)

            record_data += script_data[curr:curr+total_length]
            record_count += 1

            curr += total_length

            try:
                record = UpdateRecord.FromBinary(record_data, record_count)
            except DeferMatching as defer:
                # If we're told to defer matching, continue accumulating record_data
                # until we get a complete match.  If a partial match is available, keep track of
                # that partial match so that we can use it once the record no longer matches.

                if defer.partial_match is not None:
                    partial_match = defer.partial_match
                    match_offset = curr

                continue
            except DataError:
                if record_count > 1 and partial_match:
                    record = partial_match
                    curr = match_offset
                elif not allow_unknown:
                    raise
                elif allow_unknown and record_count > 1:
                    raise ArgumentError("A record matched an initial record subset but failed matching a subsequent addition without leaving a partial_match")
                else:
                    record = UnknownRecord(record_type, record_data[UpdateRecord.HEADER_LENGTH:])

            # Reset our record accumulator since we successfully matched one or more records
            record_count = 0
            record_data = bytearray()
            partial_match = None
            match_offset = 0

            records.append(record)

        return UpdateScript(records)

    def encode(self):
        """Encode this record into a binary blob.

        This binary blob could be parsed via a call to FromBinary().

        Returns:
            bytearray: The binary encoded script.
        """

        blob = bytearray()

        for record in self.records:
            blob += record.encode()

        header = struct.pack("<LL", self.SCRIPT_MAGIC, len(blob) + self.SCRIPT_HEADER_LENGTH)
        blob = header + blob

        sha = hashlib.sha256()
        sha.update(blob)
        hash_value = sha.digest()[:16]

        return bytearray(hash_value) + blob

    def __eq__(self, other):
        if not isinstance(other, UpdateScript):
            return False

        if len(self.records) != len(other.records):
            return False

        for record1, record2 in zip(self.records, other.records):
            if record1 != record2:
                return False

        return True

    def __ne__(self, other):
        return not self == other



import base64
import binascii
from .verifier import Verifier
from iotile.core.exceptions import ValidationError, ArgumentError


class BytesVerifier(Verifier):
    """Verify that an object is a (possibly encoded) buffer of octets

    Args:
        desc (string): A description of the verifier
        encoding (string): A supported encoding of the bytes, currently must be
            one of base64, hex or none.  If None is given, the object must be
            a bytearray type object
    """

    def __init__(self, encoding='none', desc=None):
        super(BytesVerifier, self).__init__(desc)

        known_encodings = ['base64', 'hex', 'none']
        if encoding not in known_encodings:
            raise ArgumentError("Invalid encoding for bytes", known_encodings=known_encodings)

        self.encoding = encoding

    def verify(self, obj):
        """Verify that the object conforms to this verifier's schema

        Args:
            obj (object): A python object to verify

        Returns:
            bytes or byterray: The decoded byte buffer

        Raises:
            ValidationError: If there is a problem verifying the object, a
                ValidationError is thrown with at least the reason key set indicating
                the reason for the lack of validation.
        """

        if self.encoding == 'none' and not isinstance(obj, (bytes, bytearray)):
            raise ValidationError('Byte object was not either bytes or a bytearray', type=obj.__class__.__name__)
        elif self.encoding == 'base64':
            try:
                data = base64.b64decode(obj)
                return data
            except TypeError:
                raise ValidationError("Could not decode base64 encoded bytes", obj=obj)
        elif self.encoding == 'hex':
            try:
                data = binascii.unhexlify(obj)
                return data
            except TypeError:
                raise ValidationError("Could not decode hex encoded bytes", obj=obj)

        return obj

    def format(self, indent_level, indent_size=4):
        """Format this verifier

        Returns:
            string: A formatted string
        """

        name = self.format_name('Bytes', indent_size)
        return self.wrap_lines(name, indent_level, indent_size)



"""A simple console UI that displays a list of objects with a title.

The use case for this object is that you want to create a UI that shows a list of
information where individual lines could be changing and you want to update them
in place
like this:

<TITLE>
Item 1      Value 1
Item 2      Value 2
...

You could just write a new line to the console with every changing item but that
quickly gets unusable because things change too fast and unrelated lines stomp
on each other.

LinebufferUI provides a nice way to do this.  It has the following prerequisites:

1. There must be a polling function that will return a new line.  This is
   treated as an opaque object so it may have whatever format you want.  This
   function will be called in a loop to check for changes to the items that
   are being displayed. If you return a list from this function, it will be
   treated as multiple results. This means that your object type cannot be
   anything such that isinstance(object, list) is True otherwise it will be
   misinterpretted as multiple objects.
2. You must be able to calculate a single hashable value that determines for a result
   from your polling function, what line it corresponds to.  If you wish to exclude
   some results from your polling function, simply return None from this function and
   that specific result will be dropped.
3. You must be be able to implement a function that draws each line given your object.
4. You can optionally define a function that sorts your objects, which is called to
   get the final sorted list before these are rendered to the screen.

These 3 (optionally 4) functions are called for you in order to create a nice UI
that doesn't flicker or scroll.
"""

from collections import namedtuple
import time
from future.utils import viewvalues
from past.builtins import basestring
from iotile.core.exceptions import ExternalError

LineEntry = namedtuple("LineEntry", ['text', 'id', 'sort_key', 'object'])


class LinebufferUI(object):
    """A simple console UI that displays a list of items and updates them in place."""

    def __init__(self, poll_func, id_func, draw_func, sortkey_func=None, title=None):
        self.poll_func = poll_func
        self.id_func = id_func
        self.draw_func = draw_func
        self.sortkey_func = sortkey_func

        if isinstance(title, basestring):
            self.title_func = lambda x: title
        else:
            self.title_func = title

        self.items = {}

    def run(self, refresh_interval=0.05):

        try:
            from asciimatics.screen import Screen
        except ImportError:
            raise ExternalError("You must have asciimatics installed to use LinebufferUI", suggestion="pip install iotilecore[ui]")

        Screen.wrapper(self._run_loop, arguments=[refresh_interval])

    def _update_items(self, screen):
        new_items = self.poll_func()
        if new_items is None:
            return

        if not isinstance(new_items, list):
            new_items = [new_items]

        for item in new_items:
            id_val, entry = self._parse_and_format(item, screen)
            if id_val is None:
                continue

            self.items[id_val] = entry

    def _run_loop(self, screen, refresh_interval):
        self.items = {}

        screen.clear()

        try:
            while True:
                self._update_items(screen)
                items = sorted(viewvalues(self.items), key=lambda x: x.sort_key)

                if self.title_func is not None:
                    screen.print_at(self.title_func(self.items), 0, 0)

                for line_no, item in enumerate(items):
                    screen.print_at(item.text, 0, line_no + 2)

                screen.refresh()
                time.sleep(refresh_interval)
        except KeyboardInterrupt:
            pass

    def _parse_and_format(self, item, screen):
        id_val = self.id_func(item)
        if id_val is None:
            return None, None

        text = self.draw_func(item, screen.width)

        sort_key = id_val
        if self.sortkey_func is not None:
            sort_key = self.sortkey_func(item)

        return id_val, LineEntry(text, id_val, sort_key, item)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

# Copyright (c) 2011, Bernhard Leiner
# Copyright (c) 2013-2018 Alexander Belchenko
# All rights reserved.
#
# Redistribution and use in source and binary forms,
# with or without modification, are permitted provided
# that the following conditions are met:
#
# * Redistributions of source code must retain
#   the above copyright notice, this list of conditions
#   and the following disclaimer.
# * Redistributions in binary form must reproduce
#   the above copyright notice, this list of conditions
#   and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the author nor the names
#   of its contributors may be used to endorse
#   or promote products derived from this software
#   without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Compatibility functions for python 2 and 3.

@author     Bernhard Leiner (bleiner AT gmail com)
@author     Alexander Belchenko (alexander belchenko AT gmail com)
'''

__docformat__ = "javadoc"


import sys, array


if sys.version_info[0] >= 3:
    # Python 3
    Python = 3

    def asbytes(s):
        if isinstance(s, bytes):
            return s
        return s.encode('latin1')
    def asstr(s):
        if isinstance(s, str):
            return s
        return s.decode('latin1')

    array_tobytes = getattr(array.array, "tobytes", array.array.tostring)

    IntTypes = (int,)
    StrType = str
    UnicodeType = str

    range_g = range     # range generator
    def range_l(*args): # range list
        return list(range(*args))

    def dict_keys(dikt):        # dict keys list
        return list(dikt.keys())
    def dict_keys_g(dikt):      # dict keys generator
        return dikt.keys()
    def dict_items_g(dikt):     # dict items generator
        return dikt.items()

    from io import StringIO, BytesIO

    def get_binary_stdout():
        return sys.stdout.buffer

    def get_binary_stdin():
        return sys.stdin.buffer

else:
    # Python 2
    Python = 2

    asbytes = str
    asstr = str

    array_tobytes = array.array.tostring

    IntTypes = (int, long)
    StrType = basestring
    UnicodeType = unicode

    #range_g = xrange    # range generator
    def range_g(*args):
        # we want to use xrange here but on python 2 it does not work with long ints
        try:
            return xrange(*args)
        except OverflowError:
            start = 0
            stop = 0
            step = 1
            n = len(args)
            if n == 1:
                stop = args[0]
            elif n == 2:
                start, stop = args
            elif n == 3:
                start, stop, step = args
            else:
                raise TypeError('wrong number of arguments in range_g call!')
            if step == 0:
                raise ValueError('step cannot be zero')
            if step > 0:
                def up(start, stop, step):
                    while start < stop:
                        yield start
                        start += step
                return up(start, stop, step)
            else:
                def down(start, stop, step):
                    while start > stop:
                        yield start
                        start += step
                return down(start, stop, step)

    range_l = range     # range list

    def dict_keys(dikt):        # dict keys list
        return dikt.keys()
    def dict_keys_g(dikt):      # dict keys generator
        return dikt.keys()
    def dict_items_g(dikt):     # dict items generator
        return dikt.items()

    from cStringIO import StringIO
    BytesIO = StringIO

    import os
    def _force_stream_binary(stream):
        """Force binary mode for stream on Windows."""
        if os.name == 'nt':
            f_fileno = getattr(stream, 'fileno', None)
            if f_fileno:
                fileno = f_fileno()
                if fileno >= 0:
                    import msvcrt
                    msvcrt.setmode(fileno, os.O_BINARY)
        return stream

    def get_binary_stdout():
        return _force_stream_binary(sys.stdout)

    def get_binary_stdin():
        return _force_stream_binary(sys.stdin)


"""Update records that send an RPC."""

from __future__ import (print_function, absolute_import, unicode_literals)
import struct
from binascii import hexlify
from collections import namedtuple
from future.utils import python_2_unicode_compatible
from iotile.core.exceptions import ArgumentError
from ..record import UpdateRecord, MatchQuality

RPCInfo = namedtuple("RPCInfo", ["command", "address", "response_size", "payload"])


@python_2_unicode_compatible
class SendRPCRecord(UpdateRecord):
    """Send an RPC without checking the response.

    The record causes the remote bridge script processor to send an RPC
    and not to check the contents of the result.  It still checks that the
    response size corresponds with what is expected but does not check the
    contents explicitly.

    Args:
        address (int): The address that we want to send an RPC to.
        rpc_id (int): The RPC id that we want to call
        payload (bytearray): The RPC payload that we want to send
        response_size (int): The fixed size of the expected response.  If this is None
            then the response is of a variable size and not checked explicitly.
    """

    RecordType = 3
    RecordHeaderLength = 4

    def __init__(self, address, rpc_id, payload=None, response_size=None):
        if payload is None:
            payload = bytearray()

        self.address = address
        self.rpc_id = rpc_id
        self.payload = payload
        self.variable_size = response_size is None
        self.fixed_response_size = response_size

        if self.fixed_response_size is not None and self.fixed_response_size >= (1 << 7):
            raise ArgumentError("RPC specifies a fixed response size greater than 1 << 7, which is the largest supported", size=response_size)

    def encode_contents(self):
        """Encode the contents of this update record without including a record header.

        Returns:
            bytearary: The encoded contents.
        """

        if self.variable_size:
            resp_length = 1
        else:
            resp_length = self.fixed_response_size << 1

        header = struct.pack("<HBB", self.rpc_id, self.address, resp_length)
        return bytearray(header) + self.payload

    @classmethod
    def MatchType(cls):
        """Return the record type that this record matches.

        All records must match an 8-bit record type field that is used to
        decode a binary script.  Note that multiple records may match the same
        8-bit record type if they have different levels of specificity.

        Returns:
            int: The single record type that this record matches.
        """

        return cls.RecordType

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this functon should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): The number of binary records that are included
                in record_data.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.
        """

        if record_count > 1:
            return MatchQuality.NoMatch

        # Allow other records to override us by matching against specific kinds of RPCs
        return MatchQuality.GenericMatch

    @classmethod
    def _parse_rpc_info(cls, record_data):
        if len(record_data) < SendRPCRecord.RecordHeaderLength:
            raise ArgumentError("SendRPC record is too short", header_length=4, length=len(record_data))

        cmd, address, resp_length = struct.unpack_from("<HBB", record_data)
        payload = record_data[cls.RecordHeaderLength:]

        return RPCInfo(cmd, address, resp_length, payload)

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This should be called with a binary record blob (NOT including the
        record type header) and it will decode it into a SendRPCRecord.

        Args:
            record_data (bytearray): The raw record data that we wish to parse
                into an UpdateRecord subclass NOT including its 8 byte record header.
            record_count (int): The number of records included in record_data.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.

        Returns:
            SendRPCRecord: The decoded reflash tile record.
        """

        cmd, address, resp_length, payload = cls._parse_rpc_info(record_data)

        # The first bit is 1 if we do not have a fixed length
        # The next 7 bits encode the fixed length if we do have a fixed length
        fixed_length = resp_length >> 1
        if resp_length & 0b1:
            fixed_length = None

        return cls(address, cmd, payload, fixed_length)

    def __str__(self):
        return "Call RPC 0x%X on tile at %d with payload '%s'" % (self.rpc_id, self.address, hexlify(self.payload))

    def __eq__(self, other):
        if not isinstance(other, SendRPCRecord):
            return False

        return self.rpc_id == other.rpc_id and self.address == other.address and self.payload == other.payload and self.fixed_response_size == other.fixed_response_size


@python_2_unicode_compatible
class SendErrorCheckingRPCRecord(SendRPCRecord):
    """Send an RPC and check for errors in the response.

    The record causes the remote bridge script processor to send an RPC and to
    check that the response (which must be a 32-bit error code) is 0
    indicating no error.

    Args:
        address (int): The address that we want to send an RPC to.
        rpc_id (int): The RPC id that we want to call
        payload (bytearray): The RPC payload that we want to send
        response_size (int): The fixed size of the expected response.  This must be 4
            bytes since it must return a single 32-bit error code.
    """

    RecordType = 4

    def __init__(self, address, rpc_id, payload=None, response_size=None):
        if response_size != 4:
            raise ArgumentError("Error handling RPCs must have a fixed response size of 4 bytes", response_size=response_size)

        super(SendErrorCheckingRPCRecord, self).__init__(address, rpc_id, payload, response_size)

    def __str__(self):
        return "Call RPC 0x%X on tile at %d with payload '%s' and check for errors" % (self.rpc_id, self.address, hexlify(self.payload))

    def __eq__(self, other):
        if not isinstance(other, SendErrorCheckingRPCRecord):
            return False

        return self.rpc_id == other.rpc_id and self.address == other.address and self.payload == other.payload and self.fixed_response_size == other.fixed_response_size

    @classmethod
    def parse_multiple_rpcs(cls, record_data):
        """Parse record_data into multiple error checking rpcs."""

        rpcs = []

        while len(record_data) > 0:
            total_length, record_type = struct.unpack_from("<LB3x", record_data)
            if record_type != SendErrorCheckingRPCRecord.RecordType:
                raise ArgumentError("Record set contains a record that is not an error checking RPC", record_type=record_type)

            record_contents = record_data[8: total_length]
            parsed_rpc = cls._parse_rpc_info(record_contents)
            rpcs.append(parsed_rpc)

            record_data = record_data[total_length:]

        return rpcs



"""Update script record for reflashing a tile with new firmware."""

from __future__ import (print_function, absolute_import, unicode_literals)
import struct
from future.utils import python_2_unicode_compatible
from iotile.core.exceptions import ArgumentError
from ..record import UpdateRecord, MatchQuality


@python_2_unicode_compatible
class ReflashControllerRecord(UpdateRecord):
    """Reflash an IOTile controller with updated firmware.

    This record embedds the new firmware image in the record itself and
    always targets the controller tile at address 8.

    Args:
        raw_data (bytearray): The raw binary firmware data that we should
            program
        offset (int): The absolute memory offset at which raw_data starts.
    """

    RecordType = 2
    RecordHeaderLength = 8

    def __init__(self, raw_data, offset):
        self.raw_data = raw_data
        self.offset = offset

    def encode_contents(self):
        """Encode the contents of this update record without including a record header.

        Returns:
            bytearary: The encoded contents.
        """

        header = struct.pack("<LL", self.offset, len(self.raw_data))
        return bytearray(header) + self.raw_data

    @classmethod
    def MatchType(cls):
        """Return the record type that this record matches.

        All records must match an 8-bit record type field that is used to
        decode a binary script.  Note that multiple records may match the same
        8-bit record type if they have different levels of specificity.

        Returns:
            int: The single record type that this record matches.
        """

        return ReflashControllerRecord.RecordType

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this functon should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): The number of binary records that are included
                in record_data.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.
        """

        if record_count > 1:
            return MatchQuality.NoMatch

        return MatchQuality.GenericMatch

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This should be called with a binary record blob (NOT including the
        record type header) and it will decode it into a ReflashControllerRecord.

        Args:
            record_data (bytearray): The raw record data that we wish to parse
                into an UpdateRecord subclass NOT including its 8 byte record header.
            record_count (int): The number of records included in record_data.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.

        Returns:
            ReflashControllerRecord: The decoded reflash tile record.
        """

        if len(record_data) < ReflashControllerRecord.RecordHeaderLength:
            raise ArgumentError("Record was too short to contain a full reflash record header", length=len(record_data), header_length=ReflashControllerRecord.RecordHeaderLength)

        offset, data_length = struct.unpack_from("<LL", record_data)

        bindata = record_data[ReflashControllerRecord.RecordHeaderLength:]
        if len(bindata) != data_length:
            raise ArgumentError("Embedded firmware length did not agree with actual length of embeded data", length=len(bindata), embedded_length=data_length)

        return ReflashControllerRecord(bindata, offset)

    def __eq__(self, other):
        if not isinstance(other, ReflashControllerRecord):
            return False

        return self.raw_data == other.raw_data and self.offset == other.offset

    def __str__(self):
        return "Reflash controller with %d (0x%X) bytes starting at offset %d (0x%X)" % (len(self.raw_data), len(self.raw_data), self.offset, self.offset)



"""A simple in memory kv store for testing purposes."""

# This file is copyright Arch Systems, Inc.
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.


class InMemoryKVStore(object):
    """A Key Value store based on an in memory dict

    This is intended as a drop in replacement for SQLiteKVStore.  Note that this KV store
    is not persistent across multiple process invocations so it is primarily useful for
    testing, not production use.

    Args:
        name (string): The name of the file to use as a persistent store for this KVStore.  This
            is ignored for this kv store type.
        respect_venv (bool): Make folder relative to the current virtual environment if there
            is one.  This is ignored for this kv_store type
    """

    _shared_data = {}

    def __init__(self, name, respect_venv=False):
        pass

    def get(self, key):
        """Get a value by its key.

        Args:
            key (string): The key used to store this value

        Returns:
            value (string): The value associated to the key

        Raises:
            KeyError: if the key was not found
        """

        return self._shared_data[key]

    def get_all(self):
        """Return a list of all (key, value) tuples in the kv store.

        Returns:
            list(string, string): A list of key, value pairs
        """

        return self._shared_data.items()

    def remove(self, key):
        """Remove a key from the data store.

        Args:
            key (string): The key to remove

        Raises:
            KeyError: if the key was not found
        """

        del self._shared_data[key]

    def try_get(self, key):
        """Try to get a value by its key, returning None if not found.

        Args:
            key (string): The key used to store this value

        Returns:
            value (string): The value associated to the key or None
        """

        return self._shared_data.get(key, None)

    def set(self, key, value):
        """Set the value of a key.

        Args:
            key (string): The key used to store this value
            value (string): The value to store
        """

        self._shared_data[key] = value

    def clear(self):
        """Clear all values from this kv store."""

        InMemoryKVStore._shared_data = {}



"""Shared decorators and exceptions used in virtual tiles and devices."""

import struct
import binascii
import inspect
from iotile.core.exceptions import IOTileException


class RPCNotFoundError(IOTileException):
    """Exception thrown when an RPC is not found."""
    pass


class RPCInvalidArgumentsError(IOTileException):
    """Exception thrown when an RPC with a fixed parameter format has invalid arguments."""
    pass


class RPCInvalidReturnValueError(IOTileException):
    """Exception thrown when the return value of an RPC does not conform to its known format."""
    pass


class RPCInvalidIDError(IOTileException):
    """Exception thrown when an RPC is created with an invalid RPC id."""
    pass


class TileNotFoundError(IOTileException):
    """Exception thrown when an RPC is sent to a tile that does not exist."""
    pass


def rpc(address, rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID and address.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        address (int): The address of the mock tile this RPC is for
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC
    """

    if rpc_id < 0 or rpc_id > 0xFFFF:
        raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

    def _rpc_wrapper(func):
        def _rpc_executor(self, payload):
            try:
                args = struct.unpack("<{}".format(arg_format), payload)
            except struct.error as exc:
                raise RPCInvalidArgumentsError(str(exc), arg_format=arg_format, payload=binascii.hexlify(payload))

            resp = func(self, *args)
            if resp_format is not None:
                try:
                    return struct.pack("<{}".format(resp_format), *resp)
                except struct.error as exc:
                    raise RPCInvalidReturnValueError(str(exc), resp_format=resp_format, resp=repr(resp))

            return resp

        _rpc_executor.rpc_id = rpc_id
        _rpc_executor.rpc_addr = address
        _rpc_executor.is_rpc = True
        return _rpc_executor

    return _rpc_wrapper


def tile_rpc(rpc_id, arg_format, resp_format=None):
    """Decorator to denote that a function implements an RPC with the given ID on a tile.

    The underlying function should be a member function that will take
    individual parameters after the RPC payload has been decoded according
    to arg_format.

    Arguments to the function are decoded from the 20 byte RPC argument payload according
    to arg_format, which should be a format string that can be passed to struct.unpack.

    Similarly, the function being decorated should return an iterable of results that
    will be encoded into a 20 byte response buffer by struct.pack using resp_format as
    the format string.

    The RPC will respond as if it were implemented by a tile at address ``address`` and
    the 16-bit RPC id ``rpc_id``.

    Args:
        rpc_id (int): The number of the RPC
        arg_format (string): a struct format code (without the <) for the
            parameter format for this RPC
        resp_format (string): an optional format code (without the <) for
            the response format for this RPC
    """

    return rpc(None, rpc_id, arg_format, resp_format)


class RPCDispatcher(object):
    """A simple dispatcher that can store and call RPCs."""

    def __init__(self, *args, **kwargs):
        super(RPCDispatcher, self).__init__(*args, **kwargs)
        self._rpcs = {}

        # Add any RPCs defined using decorators on this class
        for _name, value in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(value, 'is_rpc'):
                self.add_rpc(value.rpc_id, value)

    def add_rpc(self, rpc_id, func):
        """Add an RPC.

        Args:
            rpc_id (int): The ID of the RPC
            func (callable): The RPC implementation.
                The signature of callable should be callable(args) taking
                a bytes object with the argument and returning a bytes object
                with the response.
        """

        self._rpcs[rpc_id] = func

    def has_rpc(self, rpc_id):
        """Check if an RPC is defined.

        Args:
            rpc_id (int): The RPC to check

        Returns:
            bool: Whether it exists
        """

        return rpc_id in self._rpcs

    def call_rpc(self, rpc_id, payload=bytes()):
        """Call an RPC by its ID.

        Args:
            rpc_id (int): The number of the RPC
            payload (bytes): A byte string of payload parameters up to 20 bytes

        Returns:
            str: The response payload from the RPC
        """
        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if rpc_id not in self._rpcs:
            raise RPCNotFoundError("rpc_id: {}".format(rpc_id))

        return self._rpcs[rpc_id](payload)



from .verifier import Verifier
from iotile.core.exceptions import ValidationError


class LiteralVerifier(Verifier):
    """Verify that an object is equal to a literal

    Args:
        literal (object): A literal value to be checked
        desc (string): A description of the verifier
    """

    def __init__(self, literal, desc=None):
        super(LiteralVerifier, self).__init__(desc)
        self._literal = literal

    def verify(self, obj):
        """Verify that the object conforms to this verifier's schema

        Args:
            obj (object): A python object to verify

        Raises:
            ValidationError: If there is a problem verifying the dictionary, a
                ValidationError is thrown with at least the reason key set indicating
                the reason for the lack of validation.
        """

        if obj != self._literal:
            raise ValidationError("Object is not equal to literal", reason='%s is not equal to %s' % (str(obj), str(self._literal)), object=obj)

        return obj

    def format(self, indent_level, indent_size=4):
        """Format this verifier

        Returns:
            string: A formatted string
        """

        name = self.format_name('Literal', indent_size)

        if self.long_desc is not None:
            name += '\n'

        name += self.wrap_lines('value: %s\n' % str(self._literal), 1, indent_size)

        return self.wrap_lines(name, indent_level, indent_size)



"""A placeholder for unknown record types when we parse an UpdateScript."""

from __future__ import (print_function, absolute_import, unicode_literals)
import struct
from future.utils import python_2_unicode_compatible
from ..record import UpdateRecord


# pylint: disable=abstract-method, too-few-public-methods;The abstract methods should not be called, this is a placeholder class
@python_2_unicode_compatible
class UnknownRecord(UpdateRecord):
    """Reflash a tile at a specific slot with new firmware.

    This record embeds a new firmware image and targeting information for what
    slot is should be sent to.  When this record is executed on an IOTile
    device it will synchronously cause a tile to update its firmware and reset
    into the new firmware image before moving on to the next record.

    Args:
        record_type (int): The unknwn record type that we are supposed to
            process.
        raw_data (bytearray): The raw record contents.
    """

    def __init__(self, record_type, raw_data):
        self.record_type = record_type
        self.record_contents = raw_data

    def encode(self):
        """Encode this record into binary, suitable for embedded into an update script.

        This function just adds the required record header and copies the raw data
        we were passed in verbatim since we don't know what it means

        Returns:
            bytearary: The binary version of the record that could be parsed via
                a call to UpdateRecord.FromBinary()
        """

        header = struct.pack("<LB3x", len(self.record_contents) + UpdateRecord.HEADER_LENGTH, self.record_type)
        return bytearray(header) + self.record_contents

    def __eq__(self, other):
        if not isinstance(other, UnknownRecord):
            return False

        return self.record_type == other.record_type

    def __str__(self):
        return "Unknown record type %d, length: %d" % (self.record_type, len(self.record_contents))



"""Update script record for reflashing a tile with new firmware."""

from __future__ import (print_function, absolute_import, unicode_literals)
import struct
from future.utils import python_2_unicode_compatible
from iotile.core.exceptions import ArgumentError
from ..record import UpdateRecord, MatchQuality


@python_2_unicode_compatible
class ReflashTileRecord(UpdateRecord):
    """Reflash a tile at a specific slot with new firmware.

    This record embeds a new firmware image and targeting information for what
    slot is should be sent to.  When this record is executed on an IOTile device
    it will synchronously cause a tile to update its firmware and reset into the
    new firmware image before moving on to the next record.

    Args:
        slot (int): The slot number that we should target for reflashing
        raw_data (bytearray): The raw binary firmware data that we should
            program
        offset (int): The absolute memory offset at which raw_data starts.
        hardware_type (int): The hardware type of the tile that we are reflashing.
            This is currently unused and can be omitted.
    """

    RecordType = 1
    RecordHeaderLength = 20

    def __init__(self, slot, raw_data, offset, hardware_type=0):
        self.slot = slot
        self.raw_data = raw_data
        self.offset = offset
        self.hardware_type = hardware_type

    def encode_contents(self):
        """Encode the contents of this update record without including a record header.

        Returns:
            bytearary: The encoded contents.
        """

        header = struct.pack("<LL8sBxxx", self.offset, len(self.raw_data), _create_target(slot=self.slot), self.hardware_type)
        return bytearray(header) + self.raw_data

    @classmethod
    def MatchType(cls):
        """Return the record type that this record matches.

        All records must match an 8-bit record type field that is used to
        decode a binary script.  Note that multiple records may match the same
        8-bit record type if they have different levels of specificity.

        Returns:
            int: The single record type that this record matches.
        """

        return ReflashTileRecord.RecordType

    @classmethod
    def MatchQuality(cls, record_data, record_count=1):
        """Check how well this record matches the given binary data.

        This function will only be called if the record matches the type code
        given by calling MatchType() and this functon should check how well
        this record matches and return a quality score between 0 and 100, with
        higher quality matches having higher scores.  The default value should
        be MatchQuality.GenericMatch which is 50.  If this record does not
        match at all, it should return MatchQuality.NoMatch.

        Many times, only a single record type will match a given binary record
        but there are times when multiple different logical records produce
        the same type of record in a script, such as set_version and
        set_userkey both producing a call_rpc record with different RPC
        values.  The MatchQuality method is used to allow for rich decoding
        of such scripts back to the best possible record that created them.

        Args:
            record_data (bytearay): The raw record that we should check for
                a match.
            record_count (int): The number of binary records that are included
                in record_data.

        Returns:
            int: The match quality between 0 and 100.  You should use the
                constants defined in MatchQuality as much as possible.
        """

        if record_count > 1:
            return MatchQuality.NoMatch

        # Return a generic match so that someone can provide more details about
        # the firmware that loads onto a specific tile if they want.
        return MatchQuality.GenericMatch

    @classmethod
    def FromBinary(cls, record_data, record_count=1):
        """Create an UpdateRecord subclass from binary record data.

        This should be called with a binary record blob (NOT including the
        record type header) and it will decode it into a ReflashTileRecord.

        Args:
            record_data (bytearray): The raw record data that we wish to parse
                into an UpdateRecord subclass NOT including its 8 byte record header.
            record_count (int): The number of records included in record_data.

        Raises:
            ArgumentError: If the record_data is malformed and cannot be parsed.

        Returns:
            ReflashTileRecord: The decoded reflash tile record.
        """

        if len(record_data) < ReflashTileRecord.RecordHeaderLength:
            raise ArgumentError("Record was too short to contain a full reflash record header", length=len(record_data), header_length=ReflashTileRecord.RecordHeaderLength)

        offset, data_length, raw_target, hardware_type = struct.unpack_from("<LL8sB3x", record_data)

        bindata = record_data[ReflashTileRecord.RecordHeaderLength:]
        if len(bindata) != data_length:
            raise ArgumentError("Embedded firmware length did not agree with actual length of embeded data", length=len(bindata), embedded_length=data_length)

        target = _parse_target(raw_target)
        if target['controller']:
            raise ArgumentError("Invalid targetting information, you cannot reflash a controller with a ReflashTileRecord", target=target)

        return ReflashTileRecord(target['slot'], bindata, offset, hardware_type)

    def __eq__(self, other):
        if not isinstance(other, ReflashTileRecord):
            return False

        return self.slot == other.slot and self.raw_data == other.raw_data and self.offset == other.offset and self.hardware_type == other.hardware_type

    def __str__(self):
        return "Reflash slot %d with %d (0x%X) bytes starting at offset %d (0x%X) (hardware_type: %d)" % (self.slot, len(self.raw_data), len(self.raw_data), self.offset, self.offset, self.hardware_type)


_MATCH_SLOT = 1
_MATCH_CONTROLLER = 2

def _create_target(slot):
    """Create binary targetting information.

    This function implements a subset of the targetting supported
    by an embedded tileman_matchdata_t structure but this is the
    only subset that is widely used.

    Args:
        slot (int): The slot that we wish to target

    Returns:
        bytes: an 8-byte blob containing targeting information.
    """

    return struct.pack("<B6xB", slot, 1)  # 1 is kTBMatchBySlot


def _parse_target(target):
    """Parse a binary targetting information structure.

    This function only supports extracting the slot number or controller from
    the target and will raise an ArgumentError if more complicated targetting
    is desired.

    Args:
        target (bytes): The binary targetting data blob.

    Returns:
        dict: The parsed targetting data
    """

    if len(target) != 8:
        raise ArgumentError("Invalid targetting data length", expected=8, length=len(target))
    slot, match_op = struct.unpack("<B6xB", target)

    if match_op == _MATCH_CONTROLLER:
        return {'controller': True, 'slot': 0}
    elif match_op == _MATCH_SLOT:
        return {'controller': False, 'slot': slot}

    raise ArgumentError("Unsupported complex targetting specified", match_op=match_op)
